from dataclasses import dataclass
from enum import Enum
import logging
import os
from typing import Callable
import webbrowser

from mycli.constants import DOCS_URL, ISSUES_URL
from mycli.packages.sqlresult import SQLResult

try:
    if not os.environ.get('MYCLI_LLM_OFF'):
        import llm  # noqa: F401

        LLM_IMPORTED = True
    else:
        LLM_IMPORTED = False
except ImportError:
    LLM_IMPORTED = False
from pymysql.cursors import Cursor

logger = logging.getLogger(__name__)

COMMANDS: dict[str, 'SpecialCommand'] = {}
CASE_SENSITIVE_COMMANDS: set[str] = set()
CASE_INSENSITIVE_COMMANDS: set[str] = set()


class ArgType(Enum):
    NO_QUERY = 0
    PARSED_QUERY = 1
    RAW_QUERY = 2


@dataclass(frozen=True)
class SpecialCommandAlias:
    command: str
    case_sensitive: bool


@dataclass(frozen=True)
class SpecialCommand:
    handler: Callable
    command: str
    usage: str
    description: str
    arg_type: ArgType
    hidden: bool | None
    case_sensitive: bool | None
    aliases: list[SpecialCommandAlias] | None


class CommandNotFound(Exception):
    pass


class CommandVerbosity(Enum):
    SUCCINCT = "succinct"
    NORMAL = "normal"
    VERBOSE = "verbose"


def parse_special_command(sql: str) -> tuple[str, CommandVerbosity, str]:
    command, _, arg = sql.partition(" ")
    command_verbosity = CommandVerbosity.NORMAL
    if "+" in command:
        command_verbosity = CommandVerbosity.VERBOSE
    elif "-" in command:
        command_verbosity = CommandVerbosity.SUCCINCT
    command = command.strip().strip("+-")
    return (command, command_verbosity, arg.strip())


def special_command(
    command: str,
    usage: str,
    description: str,
    arg_type: ArgType = ArgType.PARSED_QUERY,
    hidden: bool = False,
    case_sensitive: bool = False,
    aliases: list[SpecialCommandAlias] | None = None,
) -> Callable:
    def wrapper(wrapped):
        register_special_command(
            wrapped,
            command,
            usage,
            description,
            arg_type=arg_type,
            hidden=hidden,
            case_sensitive=case_sensitive,
            aliases=aliases,
        )
        return wrapped

    return wrapper


def register_special_command(
    handler: Callable,
    command: str,
    usage: str,
    description: str,
    arg_type: ArgType = ArgType.PARSED_QUERY,
    hidden: bool = False,
    case_sensitive: bool = False,
    aliases: list[SpecialCommandAlias] | None = None,
) -> None:
    cmd = command.lower() if not case_sensitive else command
    COMMANDS[cmd] = SpecialCommand(
        handler,
        command,
        usage,
        description,
        arg_type=arg_type,
        hidden=hidden,
        case_sensitive=case_sensitive,
        aliases=aliases,
    )
    if case_sensitive:
        CASE_SENSITIVE_COMMANDS.add(command)
    else:
        CASE_INSENSITIVE_COMMANDS.add(command.lower())
    aliases = [] if aliases is None else aliases
    for alias in aliases:
        cmd = alias.command.lower() if not alias.case_sensitive else alias.command
        if alias.case_sensitive:
            CASE_SENSITIVE_COMMANDS.add(alias.command)
        else:
            CASE_INSENSITIVE_COMMANDS.add(alias.command.lower())
        COMMANDS[cmd] = SpecialCommand(
            handler,
            command,
            usage,
            description,
            arg_type=arg_type,
            case_sensitive=alias.case_sensitive,
            hidden=True,
            aliases=None,
        )


def execute(cur: Cursor, sql: str) -> list[SQLResult]:
    """Execute a special command and return the results. If the special command
    is not supported a CommandNotFound will be raised.
    """
    command, command_verbosity, arg = parse_special_command(sql)

    if (command not in CASE_SENSITIVE_COMMANDS) and (command.lower() not in CASE_INSENSITIVE_COMMANDS):
        raise CommandNotFound(f'Command not found: {command}')

    try:
        special_cmd = COMMANDS[command]
    except KeyError as exc:
        special_cmd = COMMANDS[command.lower()]
        if special_cmd.case_sensitive:
            raise CommandNotFound(f'Command not found: {command}') from exc

    # "help <SQL KEYWORD> is a special case. We want built-in help, not
    # mycli help here.
    if command.lower() == "help" and arg:
        return show_keyword_help(cur=cur, arg=arg)

    if special_cmd.arg_type == ArgType.NO_QUERY:
        return special_cmd.handler()
    elif special_cmd.arg_type == ArgType.PARSED_QUERY:
        return special_cmd.handler(cur=cur, arg=arg, command_verbosity=(command_verbosity == CommandVerbosity.VERBOSE))
    elif special_cmd.arg_type == ArgType.RAW_QUERY:
        return special_cmd.handler(cur=cur, query=sql)

    raise CommandNotFound(f"Command type not found: {command}")


@special_command(
    "help",
    "help [term]",
    "Show this table, or search for help on a term.",
    arg_type=ArgType.NO_QUERY,
    aliases=[SpecialCommandAlias("\\?", case_sensitive=False), SpecialCommandAlias("?", case_sensitive=False)],
)
def show_help(*_args) -> list[SQLResult]:
    header = ["Command", "Shortcut", "Usage", "Description"]
    result = []

    for _, value in sorted(COMMANDS.items(), key=lambda x: str.casefold(x[0])):
        if value.hidden:
            continue
        if value.aliases:
            shortcut = value.aliases[0].command
        else:
            shortcut = None
        result.append((value.command, shortcut, value.usage, value.description))
    return [SQLResult(header=header, rows=result, postamble=f'Docs index — {DOCS_URL}')]


def _show_special_help(keyword: str) -> list[SQLResult]:
    header = ['name', 'description', 'example']
    command = COMMANDS[keyword]
    description = '\n'.join([command.usage or '', command.description])
    rows = [(keyword, description, '')]
    return [SQLResult(header=header, rows=rows)]


def _show_mysql_help(cur: Cursor, keyword: str) -> list[SQLResult]:
    """
    Call the built-in "show <keyword>", to display help for an SQL keyword.
    :param cur: cursor
    :param arg: string
    :return: list
    """
    query = 'help %s'
    logger.debug(query)
    cur.execute(query, keyword)
    if cur.description and cur.rowcount > 0:
        header = [x[0] for x in cur.description]
        return [SQLResult(header=header, rows=cur)]
    logger.debug(query)
    cur.execute(query, (f'%{keyword}%',))
    if cur.description and cur.rowcount > 0:
        header = [x[0] for x in cur.description]
        return [SQLResult(preamble='Similar terms:', header=header, rows=cur)]
    else:
        return [SQLResult(status=f'No help found for "{keyword}".')]


def show_keyword_help(cur: Cursor, arg: str) -> list[SQLResult]:
    keyword = arg.strip().strip('"').strip("'").rstrip('+-')

    if keyword in CASE_SENSITIVE_COMMANDS:
        return _show_special_help(keyword)
    elif keyword.lower() in CASE_INSENSITIVE_COMMANDS:
        return _show_special_help(keyword.lower())

    return _show_mysql_help(cur, keyword)


@special_command('\\bug', '\\bug', 'File a bug on GitHub.', arg_type=ArgType.NO_QUERY)
def file_bug(*_args) -> list[SQLResult]:
    webbrowser.open_new_tab(ISSUES_URL)
    return [SQLResult(status=f'{ISSUES_URL} — press "New Issue"')]


@special_command(
    "exit",
    "exit",
    "Exit.",
    arg_type=ArgType.NO_QUERY,
    aliases=[SpecialCommandAlias("\\q", case_sensitive=False)],
)
@special_command(
    "quit",
    "quit",
    "Quit.",
    arg_type=ArgType.NO_QUERY,
    aliases=[SpecialCommandAlias("\\q", case_sensitive=False)],
)
def quit_(*_args):
    raise EOFError


@special_command(
    "\\edit",
    "<query>\\edit | \\edit <filename>",
    "Edit query with editor (uses $VISUAL or $EDITOR).",
    arg_type=ArgType.NO_QUERY,
    case_sensitive=True,
    aliases=[SpecialCommandAlias("\\e", case_sensitive=True)],
)
@special_command(
    "\\clip",
    "<query>\\clip",
    "Copy query to the system clipboard.",
    arg_type=ArgType.NO_QUERY,
    case_sensitive=True,
)
@special_command(
    "\\G",
    "<query>\\G",
    "Display query results vertically.",
    arg_type=ArgType.NO_QUERY,
    case_sensitive=True,
)
@special_command(
    "\\g",
    "<query>\\g",
    "Display query results (mnemonic: go).",
    arg_type=ArgType.NO_QUERY,
    case_sensitive=True,
)
def stub():
    raise NotImplementedError


if LLM_IMPORTED:

    @special_command(
        "\\llm",
        "\\llm [arguments]",
        "Interrogate an LLM.  See \"\\llm help\".",
        arg_type=ArgType.RAW_QUERY,
        case_sensitive=True,
        aliases=[SpecialCommandAlias("\\ai", case_sensitive=True)],
    )
    def llm_stub():
        raise NotImplementedError



from __future__ import annotations

import itertools
import shutil
from typing import Any, cast

import click
from configobj import ConfigObj
import prompt_toolkit
from prompt_toolkit.formatted_text import ANSI, FormattedText, to_plain_text
import pytest

from mycli import output as output_module
from mycli.output import OutputMixin
from mycli.packages.sqlresult import SQLResult
from test.utils import DummyFormatter, FakeCursorBase, make_bare_mycli  # type: ignore[attr-defined]


def test_output_timing_logs_and_prints_with_default_style(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    logged: list[Any] = []
    printed: list[tuple[Any, Any]] = []
    cli.log_output = lambda value: logged.append(value)  # type: ignore[assignment]
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: printed.append((text, style)))

    OutputMixin.output_timing(cli, '0.12 sec')

    assert logged == ['0.12 sec']
    assert to_plain_text(printed[0][0]) == '0.12 sec'
    assert list(printed[0][0])[0][0].strip() == 'class:output.timing'
    assert printed[0][1] == cli.ptoolkit_style


def test_output_timing_uses_warning_style(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.log_output = lambda value: None  # type: ignore[assignment]
    printed: list[Any] = []
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: printed.append(text))

    OutputMixin.output_timing(cli, '0.34 sec', is_warnings_style=True)

    assert list(printed[0])[0][0].strip() == 'class:warnings.timing'


def test_log_query_and_log_output_write_plain_text(tmp_path) -> None:
    cli = make_bare_mycli()
    logfile = tmp_path / 'audit.log'

    with logfile.open('w+', encoding='utf-8') as handle:
        cli.logfile = handle
        OutputMixin.log_query(cli, 'select 1')
        OutputMixin.log_output(cli, ANSI('\x1b[31mhello\x1b[0m'))
        handle.seek(0)
        contents = handle.read()

    assert 'select 1' in contents
    assert 'hello' in contents
    assert '\x1b[31m' not in contents


def test_log_output_ignores_missing_logfile() -> None:
    cli = make_bare_mycli()
    cli.logfile = None

    OutputMixin.log_output(cli, 'nothing to write')


def test_echo_logs_and_prints(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    logged: list[str] = []
    printed: list[tuple[str, dict[str, Any]]] = []
    cli.log_output = lambda value: logged.append(value)  # type: ignore[assignment]
    monkeypatch.setattr(click, 'secho', lambda value, **kwargs: printed.append((value, kwargs)))

    OutputMixin.echo(cli, 'message', fg='red')

    assert logged == ['message']
    assert printed == [('message', {'fg': 'red'})]


def test_get_output_margin_renders_prompt_once_and_counts_status_lines(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_lines = 0
    cli.prompt_format = 'ignored'
    cli.prompt_session = None
    cli.get_reserved_space = lambda: 2  # type: ignore[assignment]
    monkeypatch.setattr(output_module.repl_mode, 'render_prompt_string', lambda *_args: FormattedText([('', 'one\ntwo')]))
    monkeypatch.setattr(output_module.special, 'is_timing_enabled', lambda: True)

    margin = OutputMixin.get_output_margin(cli, 'ok\nwarning')

    assert margin == 7
    assert cli.prompt_lines == 2


def test_output_writes_lines_sinks_and_status(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.explicit_pager = False
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    logged: list[Any] = []
    tee: list[str] = []
    once: list[str] = []
    pipe_once: list[str] = []
    printed_lines: list[str] = []
    printed_status: list[Any] = []
    cli.log_output = lambda value: logged.append(value)  # type: ignore[assignment]
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: tee.append(value))
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: once.append(value))
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: pipe_once.append(value))
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_pager_enabled', lambda: False)
    monkeypatch.setattr(click, 'secho', lambda value, **_kwargs: printed_lines.append(value))
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: printed_status.append(text))

    OutputMixin.output(cli, itertools.chain(['row 1', 'row 2']), SQLResult(status='done'))

    assert logged == ['row 1', 'row 2', 'done']
    assert tee == ['row 1', 'row 2']
    assert once == ['row 1', 'row 2']
    assert pipe_once == ['row 1', 'row 2']
    assert printed_lines == ['row 1', 'row 2']
    assert to_plain_text(printed_status[0]) == 'done'
    assert list(printed_status[0])[0][0].strip() == 'class:output.status'


def test_output_uses_warning_status_style(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    printed_status: list[Any] = []
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: printed_status.append(text))

    OutputMixin.output(cli, itertools.chain([]), SQLResult(status='warning'), is_warnings_style=True)

    assert list(printed_status[0])[0][0].strip() == 'class:warnings.status'


def test_output_sends_buffer_to_pager_when_pager_is_explicit(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.prompt_session = None
    cli.explicit_pager = True
    cli.log_output = lambda value: None  # type: ignore[assignment]
    cli.get_output_margin = lambda status=None: 1  # type: ignore[assignment]
    paged_lines: list[str] = []
    monkeypatch.setattr(output_module.special, 'write_tee', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'write_pipe_once', lambda value: None)
    monkeypatch.setattr(output_module.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(output_module.special, 'is_pager_enabled', lambda: True)
    monkeypatch.setattr(click, 'echo_via_pager', lambda values: paged_lines.extend(list(values)))
    monkeypatch.setattr(prompt_toolkit, 'print_formatted_text', lambda text, style=None: None)

    OutputMixin.output(cli, itertools.chain(['row 1', 'row 2']), SQLResult())

    assert paged_lines == ['row 1\n', 'row 2\n']


def test_configure_pager_prefers_my_cnf_pager_and_sets_less(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.my_cnf = ConfigObj({'client': {'pager': 'my-pager'}})
    cli.config = ConfigObj({'main': {'pager': 'config-pager', 'enable_pager': 'True'}})
    cli.read_my_cnf = lambda cnf, keys: {'pager': 'my-pager', 'skip-pager': None}  # type: ignore[assignment]
    pager_calls: list[str] = []
    disabled: list[bool] = []
    monkeypatch.delenv('LESS', raising=False)
    monkeypatch.setattr(output_module.special, 'set_pager', lambda value: pager_calls.append(value))
    monkeypatch.setattr(output_module.special, 'disable_pager', lambda: disabled.append(True))

    OutputMixin.configure_pager(cli)

    assert pager_calls == ['my-pager']
    assert disabled == []
    assert cli.explicit_pager is True
    assert output_module.os.environ['LESS'] == '-RXF'


def test_configure_pager_disables_when_skip_pager_is_set(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.my_cnf = ConfigObj({'client': {}})
    cli.config = ConfigObj({'main': {'pager': '', 'enable_pager': 'True'}})
    cli.read_my_cnf = lambda cnf, keys: {'pager': None, 'skip-pager': '1'}  # type: ignore[assignment]
    disabled: list[bool] = []
    monkeypatch.setattr(output_module.special, 'set_pager', lambda value: None)
    monkeypatch.setattr(output_module.special, 'disable_pager', lambda: disabled.append(True))

    OutputMixin.configure_pager(cli)

    assert cli.explicit_pager is False
    assert disabled == [True]


def test_format_sqlresult_uses_redirect_formatter_and_appends_preamble_postamble() -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    cli.redirect_formatter = DummyFormatter()
    result = SQLResult(preamble='before', header=['id'], rows=[(1,)], postamble='after')

    formatted = list(OutputMixin.format_sqlresult(cli, result, is_redirected=True))

    assert formatted == ['before', 'plain output', 'after']
    assert cli.main_formatter.calls == []
    assert cli.redirect_formatter.calls


def test_format_sqlresult_for_cursor_sets_column_types_and_alignment(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    monkeypatch.setattr(output_module, 'Cursor', FakeCursorBase)
    rows = FakeCursorBase(rows=[(1, 'name')], rowcount=1, description=[('id', 3), ('name', 253)])
    result = SQLResult(header=['id', 'name'], rows=cast(Any, rows))

    assert list(OutputMixin.format_sqlresult(cli, result, numeric_alignment='left')) == ['plain output']

    _, kwargs = cli.main_formatter.calls[-1]
    assert kwargs['column_types'] == [int, str]
    assert kwargs['colalign'] == ['left', 'left']


def test_format_sqlresult_switches_to_vertical_when_first_line_is_too_wide() -> None:
    cli = make_bare_mycli()
    cli.main_formatter = DummyFormatter()
    result = SQLResult(header=['id'], rows=[(1,)])

    assert list(OutputMixin.format_sqlresult(cli, result, max_width=2)) == ['vertical output']


def test_get_reserved_space_caps_ratio(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_bare_mycli()
    monkeypatch.setattr(shutil, 'get_terminal_size', lambda *args, **kwargs: (120, 40))

    assert OutputMixin.get_reserved_space(cli) == 8



# type: ignore

from io import StringIO
import re
import sys
import textwrap

import pexpect


def expect_exact(context, expected, timeout):
    timedout = False
    try:
        context.cli.expect_exact(expected, timeout=timeout)
    except pexpect.TIMEOUT:
        timedout = True
    if timedout:
        # Strip color codes out of the output.
        actual = re.sub(r"\x1b\[([0-9A-Za-z;?])+[m|K]?", "", context.cli.before)
        raise Exception(
            textwrap.dedent(
                f"""\
                Expected:
                ---
                {expected!r}
                ---
                Actual:
                ---
                {actual!r}
                ---
                Full log:
                ---
                {context.logfile.getvalue()!r}
                ---
                """
            )
        )


def expect_pager(context, expected, timeout):
    expect_exact(context, f"{context.conf['pager_boundary']}\r\n{expected}{context.conf['pager_boundary']}\r\n", timeout=timeout)


def run_cli(context, run_args=None, exclude_args=None):
    """Run the process using pexpect."""
    run_args = run_args or {}
    rendered_args = []
    exclude_args = set(exclude_args) if exclude_args else set()

    conf = dict(**context.conf)
    conf.update(run_args)

    def add_arg(name, key, value):
        if name not in exclude_args:
            if value is not None:
                rendered_args.extend((key, value))
            else:
                rendered_args.append(key)

    if conf.get("host", None):
        add_arg("host", "-h", conf["host"])
    if conf.get("user", None):
        add_arg("user", "-u", conf["user"])
    if conf.get("pass", None):
        add_arg("pass", "-p", conf["pass"])
    if conf.get("port", None):
        add_arg("port", "-P", str(conf["port"]))
    if conf.get("dbname", None):
        add_arg("dbname", "-D", conf["dbname"])
    if conf.get("defaults-file", None):
        add_arg("defaults_file", "--defaults-file", conf["defaults-file"])
    if conf.get("myclirc", None):
        add_arg("myclirc", "--myclirc", conf["myclirc"])
    if conf.get("login_path"):
        add_arg("login_path", "--login-path", conf["login_path"])

    for arg_name, arg_value in conf.items():
        if arg_name.startswith("-"):
            add_arg(arg_name, arg_name, arg_value)

    try:
        cli_cmd = context.conf["cli_command"]
    except KeyError:
        cli_cmd = f'{sys.executable} -c "import coverage ; coverage.process_startup(); import mycli.main; mycli.main.click_entrypoint()"'

    cmd_parts = [cli_cmd] + rendered_args
    cmd = " ".join(cmd_parts)
    context.cli = pexpect.spawnu(cmd, cwd=context.package_root)
    context.logfile = StringIO()
    context.cli.logfile = context.logfile
    context.exit_sent = False
    context.currentdb = context.conf["dbname"]


def wait_prompt(context, prompt=None):
    """Make sure prompt is displayed."""
    if prompt is None:
        user = context.conf["user"]
        host = context.conf["host"]
        dbname = context.currentdb
        prompt = (f"{user}@{host}:{dbname}>",)
    expect_exact(context, prompt, timeout=5)
    context.atprompt = True



# type: ignore

from types import SimpleNamespace

import pytest
import sqlparse

from mycli.packages import completion_engine, special
from mycli.packages.completion_engine import (
    _aliases,
    _build_suggest_context,
    _charset_suggestion,
    _emit_binary_or_comma,
    _emit_blank_token,
    _emit_character_set,
    _emit_collation,
    _emit_column_for_tables,
    _emit_database,
    _emit_lparen,
    _emit_none_token,
    _emit_nothing,
    _emit_on,
    _emit_procedure,
    _emit_relation_like,
    _emit_relation_name,
    _emit_select_like,
    _emit_show,
    _emit_star,
    _emit_to,
    _emit_user,
    _emit_where_token,
    _enum_value_suggestion,
    _find_doubled_backticks,
    _is_single_or_double_quoted,
    _is_where_or_having,
    _keyword_and_special_suggestions,
    _keyword_suggestions,
    _normalize_token_value,
    _parent_name,
    _parse_suggestion_statement,
    _tables,
    _token_is_binary_or_comma,
    _token_is_blank,
    _token_is_lparen,
    _token_is_none,
    _token_is_relation_keyword,
    _token_value_is,
    _tokens_wo_space,
    _word_starts_with_digit_or_dot,
    _word_starts_with_quote,
    identifies,
    is_inside_quotes,
    suggest_based_on_last_token,
    suggest_special,
    suggest_type,
)


def sorted_dicts(dicts):
    """input is a list of dicts."""
    return sorted(tuple(x.items()) for x in dicts)


def flattened_tokens(text):
    return list(sqlparse.parse(text)[0].flatten())


def value_tokens(*values):
    return [SimpleNamespace(value=value) for value in values]


def empty_identifier():
    return SimpleNamespace(get_parent_name=lambda: None)


def last_non_whitespace_token(text):
    parsed = sqlparse.parse(text)[0]
    return parsed.token_prev(len(parsed.tokens) - 1)[1]


def test_select_suggests_cols_with_visible_table_scope():
    suggestions = suggest_type("SELECT  FROM tabl", "SELECT ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["tabl"]},
        {"type": "column", "tables": [(None, "tabl", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_select_suggests_cols_with_qualified_table_scope():
    suggestions = suggest_type("SELECT  FROM sch.tabl", "SELECT ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["tabl"]},
        {"type": "column", "tables": [("sch", "tabl", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM tabl WHERE ",
        "SELECT * FROM tabl WHERE (",
        "SELECT * FROM tabl WHERE bar OR ",
        "SELECT * FROM tabl WHERE foo = 1 AND ",
        "SELECT * FROM tabl WHERE (bar > 10 AND ",
        "SELECT * FROM tabl WHERE (bar AND (baz OR (qux AND (",
        "SELECT * FROM tabl WHERE 10 < ",
        "SELECT * FROM tabl WHERE foo BETWEEN ",
        "SELECT * FROM tabl WHERE foo BETWEEN foo AND ",
    ],
)
def test_where_suggests_columns_functions(expression):
    suggestions = suggest_type(expression, expression)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["tabl"]},
        {"type": "column", "tables": [(None, "tabl", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_where_equals_suggests_enum_values_first():
    expression = "SELECT * FROM tabl WHERE foo = "
    suggestions = suggest_type(expression, expression)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "enum_value", "tables": [(None, "tabl", None)], "column": "foo", "parent": None},
        {"type": "alias", "aliases": ["tabl"]},
        {"type": "column", "tables": [(None, "tabl", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_enum_value_suggestion_returns_none_without_equals_context():
    expression = 'SELECT * FROM tabl WHERE foo'
    suggestion = _enum_value_suggestion(expression, expression)
    assert suggestion is None


def test_enum_value_suggestion_returns_column_and_tables():
    expression = 'SELECT * FROM tabl WHERE foo = '
    suggestion = _enum_value_suggestion(expression, expression)
    assert suggestion == {
        'type': 'enum_value',
        'tables': [(None, 'tabl', None)],
        'column': 'foo',
        'parent': None,
    }


def test_enum_value_suggestion_handles_qualified_backticked_identifier():
    expression = 'SELECT * FROM sch.tabl WHERE `tabl`.`foo` = '
    suggestion = _enum_value_suggestion(expression, expression)
    assert suggestion == {
        'type': 'enum_value',
        'tables': [('sch', 'tabl', None)],
        'column': '`foo`',
        'parent': '`tabl`',
    }


def test_enum_value_suggestion_returns_none_inside_quotes():
    full_text = 'SELECT * FROM tabl WHERE "foo = '
    text_before_cursor = 'SELECT * FROM tabl WHERE "foo = '
    suggestion = _enum_value_suggestion(text_before_cursor, full_text)
    assert suggestion is None


@pytest.mark.parametrize(
    ('tokens', 'expected'),
    [
        (value_tokens('character', 'set'), [{'type': 'character_set'}]),
        (value_tokens('x', 'character', 'set', ' '), [{'type': 'character_set'}]),
        (value_tokens('collate'), [{'type': 'collation'}]),
        (value_tokens('select', 'foo'), None),
    ],
)
def test_charset_suggestion(tokens, expected):
    assert _charset_suggestion(tokens) == expected


def test_keyword_suggestions():
    assert _keyword_suggestions() == [{'type': 'keyword'}]


def test_keyword_and_special_suggestions():
    assert _keyword_and_special_suggestions() == [{'type': 'keyword'}, {'type': 'special'}]


def test_parse_suggestion_statement_returns_statement_and_nonspace_tokens():
    tokens_wo_space = _tokens_wo_space('select  1')
    assert [token.value for token in tokens_wo_space] == ['select', '1']


def test_parse_suggestion_statement_raises_type_error_for_invalid_input_type():
    with pytest.raises(TypeError):
        _parse_suggestion_statement(None)  # type: ignore[arg-type]


def test_normalize_token_value_handles_string():
    assert _normalize_token_value('SELECT') == 'select'


def test_normalize_token_value_handles_none():
    assert _normalize_token_value(None) is None


def test_normalize_token_value_handles_plain_token():
    token = SimpleNamespace(value='SHOW')
    assert _normalize_token_value(token) == 'show'


def test_normalize_token_value_handles_comparison_token():
    comparison = sqlparse.parse('a.id = d.')[0].tokens[0]
    assert _normalize_token_value(comparison) == 'd.'


def test_build_suggest_context_populates_fields():
    identifier = empty_identifier()
    context = _build_suggest_context(
        'SHOW',
        'show ',
        None,
        'show ',
        identifier,
    )

    assert context.token == 'SHOW'
    assert context.token_value == 'show'
    assert context.text_before_cursor == 'show '
    assert context.word_before_cursor is None
    assert context.full_text == 'show '
    assert context.identifier is identifier
    assert str(context.parsed_cb()) == 'show '
    assert [token.value for token in context.tokens_wo_space_cb()] == ['show']


def test_build_suggest_context_handles_none_token():
    context = _build_suggest_context(
        None,
        '',
        None,
        '',
        empty_identifier(),
    )

    assert context.token is None
    assert context.token_value is None
    assert str(context.parsed_cb()) == ''
    assert context.tokens_wo_space_cb() == []


@pytest.mark.parametrize(
    ('text_before_cursor', 'expected'),
    [
        ("select 'foo", True),
        ('select "foo', True),
        ('select `foo', False),
        ('select foo', False),
    ],
)
def test_is_single_or_double_quoted(text_before_cursor, expected):
    context = _build_suggest_context(
        None,
        text_before_cursor,
        None,
        text_before_cursor,
        empty_identifier(),
    )
    assert _is_single_or_double_quoted(context) is expected


def test_parent_name_returns_identifier_parent():
    identifier = SimpleNamespace(get_parent_name=lambda: 'sch')
    context = _build_suggest_context(None, '', None, '', identifier)
    assert _parent_name(context) == 'sch'


def test_parent_name_returns_empty_list_without_parent():
    context = _build_suggest_context(None, '', None, '', empty_identifier())
    assert _parent_name(context) == []


def test_tables_returns_extracted_tables_from_full_text():
    full_text = 'SELECT * FROM abc a, sch.def d'
    context = _build_suggest_context(None, '', None, full_text, empty_identifier())
    assert _tables(context) == [
        (None, 'abc', 'a'),
        ('sch', 'def', 'd'),
    ]


def test_aliases_prefers_alias_and_falls_back_to_table_name():
    tables = [
        (None, 'abc', 'a'),
        ('sch', 'def', ''),
    ]
    assert _aliases(tables) == ['a', 'def']


@pytest.mark.parametrize(
    ('word_before_cursor', 'expected'),
    [
        ('9foo', True),
        ('.foo', True),
        ('foo', False),
        (None, False),
    ],
)
def test_word_starts_with_digit_or_dot(word_before_cursor, expected):
    context = _build_suggest_context(
        None,
        '',
        word_before_cursor,
        '',
        empty_identifier(),
    )
    assert _word_starts_with_digit_or_dot(context) is expected


@pytest.mark.parametrize(
    ('word_before_cursor', 'expected'),
    [
        ("'foo", True),
        ('"foo', True),
        ('foo', False),
        (None, False),
    ],
)
def test_word_starts_with_quote(word_before_cursor, expected):
    context = _build_suggest_context(
        None,
        '',
        word_before_cursor,
        '',
        empty_identifier(),
    )
    assert _word_starts_with_quote(context) is expected


def test_token_is_none_true_for_none_token():
    context = _build_suggest_context(None, '', None, '', empty_identifier())
    assert _token_is_none(context) is True


def test_token_is_none_false_for_non_none_token():
    context = _build_suggest_context('select', '', None, '', empty_identifier())
    assert _token_is_none(context) is False


@pytest.mark.parametrize(
    ('token', 'expected'),
    [
        ('', True),
        ('select', False),
        (None, True),
    ],
)
def test_token_is_blank(token, expected):
    context = _build_suggest_context(token, '', None, '', empty_identifier())
    assert _token_is_blank(context) is expected


@pytest.mark.parametrize(
    ('token', 'values', 'expected'),
    [
        ('select', ('select', 'where'), True),
        ('show', ('select', 'where'), False),
        (None, ('select',), False),
    ],
)
def test_token_value_is(token, values, expected):
    context = _build_suggest_context(token, '', None, '', empty_identifier())
    assert _token_value_is(context, *values) is expected


@pytest.mark.parametrize(
    ('token', 'expected'),
    [
        ('(', True),
        ('any(', True),
        ('select', False),
        (None, False),
    ],
)
def test_token_is_lparen(token, expected):
    context = _build_suggest_context(token, '', None, '', empty_identifier())
    assert _token_is_lparen(context) is expected


@pytest.mark.parametrize(
    ('token', 'text_before_cursor', 'full_text', 'expected'),
    [
        (last_non_whitespace_token('SELECT * FROM foo JOIN '), 'SELECT * FROM foo JOIN ', 'SELECT * FROM foo JOIN ', True),
        ('from', 'from ', 'from ', True),
        ('truncate', 'truncate ', 'truncate ', True),
        ('like', 'like ', 'create table new like ', True),
        ('like', 'like ', 'select * from foo like ', False),
        ('select', 'select ', 'select ', False),
    ],
)
def test_token_is_relation_keyword(token, text_before_cursor, full_text, expected):
    context = _build_suggest_context(token, text_before_cursor, None, full_text, empty_identifier())
    assert _token_is_relation_keyword(context) is expected


@pytest.mark.parametrize(
    ('token', 'expected'),
    [
        (',', True),
        ('=', True),
        ('and', True),
        ('select', False),
        (None, False),
    ],
)
def test_token_is_binary_or_comma(token, expected):
    context = _build_suggest_context(token, '', None, '', empty_identifier())
    assert _token_is_binary_or_comma(context) is expected


def test_emit_none_token():
    context = _build_suggest_context(None, '', None, '', empty_identifier())
    assert _emit_none_token(context) == [{'type': 'keyword'}]


def test_emit_blank_token():
    context = _build_suggest_context('', '', None, '', empty_identifier())
    assert _emit_blank_token(context) == [{'type': 'keyword'}, {'type': 'special'}]


def test_emit_star():
    context = _build_suggest_context('*', '', None, '', empty_identifier())
    assert _emit_star(context) == [{'type': 'keyword'}]


def test_emit_lparen_exists_where():
    text = 'SELECT * FROM foo WHERE EXISTS ('
    context = _build_suggest_context('(', text, None, text, empty_identifier())
    assert _emit_lparen(context) == [{'type': 'keyword'}]


def test_emit_lparen_join_using():
    text = 'select * from abc inner join def using ('
    context = _build_suggest_context('(', text, None, text, empty_identifier())
    assert _emit_lparen(context) == [{'type': 'column', 'tables': [(None, 'abc', None), (None, 'def', None)], 'drop_unique': True}]


def test_emit_lparen_show():
    text = 'SHOW ('
    context = _build_suggest_context('(', text, None, text, empty_identifier())
    assert _emit_lparen(context) == [{'type': 'show'}]


def test_emit_lparen_function_argument_list():
    text = 'SELECT MAX('
    full_text = 'SELECT MAX( FROM tbl'
    context = _build_suggest_context('(', text, None, full_text, empty_identifier())
    assert _emit_lparen(context) == [{'type': 'column', 'tables': [(None, 'tbl', None)]}]


def test_emit_procedure():
    context = _build_suggest_context('call', '', None, '', empty_identifier())
    assert _emit_procedure(context) == [{'type': 'procedure', 'schema': []}]


def test_emit_character_set():
    context = _build_suggest_context('set', '', None, '', empty_identifier())
    assert _emit_character_set(context) == [{'type': 'character_set'}]


def test_emit_column_for_tables():
    full_text = 'SELECT * FROM abc a, sch.def d'
    context = _build_suggest_context('select', '', None, full_text, empty_identifier())
    assert _emit_column_for_tables(context) == [
        {
            'type': 'column',
            'tables': [
                (None, 'abc', 'a'),
                ('sch', 'def', 'd'),
            ],
        }
    ]


def test_emit_nothing():
    context = _build_suggest_context('as', '', None, '', empty_identifier())
    assert _emit_nothing(context) == []


def test_emit_show():
    context = _build_suggest_context('show', '', None, '', empty_identifier())
    assert _emit_show(context) == [{'type': 'show'}]


def test_emit_to_for_change_statement():
    text = 'change master to '
    context = _build_suggest_context('to', text, None, text, empty_identifier())
    assert _emit_to(context) == [{'type': 'change'}]


def test_emit_to_for_non_change_statement():
    text = 'grant all on db.* to '
    context = _build_suggest_context('to', text, None, text, empty_identifier())
    assert _emit_to(context) == [{'type': 'user'}]


def test_emit_user():
    context = _build_suggest_context('user', '', None, '', empty_identifier())
    assert _emit_user(context) == [{'type': 'user'}]


def test_emit_collation():
    context = _build_suggest_context('collate', '', None, '', empty_identifier())
    assert _emit_collation(context) == [{'type': 'collation'}]


@pytest.mark.xfail
def test_emit_select_like_with_parent_filters_tables():
    identifier = SimpleNamespace(get_parent_name=lambda: 't1')
    text = 'SELECT t1.'
    full_text = 'SELECT t1. FROM tabl1 t1, tabl2 t2'
    context = _build_suggest_context('select', text, None, full_text, identifier)
    assert sorted_dicts(_emit_select_like(context)) == sorted_dicts([
        {'type': 'column', 'tables': [(None, 'tabl1', 't1')]},
        # xfail because these are also currently returned
        # {'type': 'table', 'schema': 't1'},
        # {'type': 'view', 'schema': 't1'},
        # {'type': 'function', 'schema': 't1'},
    ])


def test_emit_select_like_inside_backticks_adds_keyword():
    text = 'SELECT `a'
    full_text = 'SELECT `a FROM tabl'
    context = _build_suggest_context('select', text, None, full_text, empty_identifier())
    assert sorted_dicts(_emit_select_like(context)) == sorted_dicts([
        {'type': 'column', 'tables': [(None, 'tabl', None)]},
        {'type': 'function', 'schema': []},
        {'type': 'alias', 'aliases': ['tabl']},
        {'type': 'keyword'},
    ])


def test_emit_select_like_default():
    text = 'SELECT '
    full_text = 'SELECT  FROM tabl'
    context = _build_suggest_context('select', text, None, full_text, empty_identifier())
    assert sorted_dicts(_emit_select_like(context)) == sorted_dicts([
        {'type': 'column', 'tables': [(None, 'tabl', None)]},
        {'type': 'function', 'schema': []},
        {'type': 'introducer'},
        {'type': 'alias', 'aliases': ['tabl']},
    ])


def test_emit_relation_like_with_schema_parent():
    identifier = SimpleNamespace(get_parent_name=lambda: 'sch')
    text = 'INSERT INTO sch.'
    context = _build_suggest_context('into', text, None, text, identifier)
    assert sorted_dicts(_emit_relation_like(context)) == sorted_dicts([
        {'type': 'table', 'schema': 'sch'},
        {'type': 'view', 'schema': 'sch'},
    ])


def test_emit_relation_like_join_adds_database_and_join_flag():
    text = 'SELECT * FROM foo JOIN '
    token = last_non_whitespace_token(text)
    context = _build_suggest_context(token, text, None, text, empty_identifier())
    assert sorted_dicts(_emit_relation_like(context)) == sorted_dicts([
        {'type': 'database'},
        {'type': 'table', 'schema': [], 'join': True},
        {'type': 'view', 'schema': []},
    ])


def test_emit_relation_like_truncate_omits_view():
    text = 'TRUNCATE '
    context = _build_suggest_context('truncate', text, None, text, empty_identifier())
    assert sorted_dicts(_emit_relation_like(context)) == sorted_dicts([
        {'type': 'database'},
        {'type': 'table', 'schema': []},
    ])


def test_emit_relation_name_with_schema_parent():
    identifier = SimpleNamespace(get_parent_name=lambda: 'sch')
    context = _build_suggest_context('table', '', None, '', identifier)
    assert _emit_relation_name(context) == [{'type': 'table', 'schema': 'sch'}]


def test_emit_relation_name_without_schema_parent():
    context = _build_suggest_context('view', '', None, '', empty_identifier())
    assert _emit_relation_name(context) == [{'type': 'schema'}, {'type': 'view', 'schema': []}]


@pytest.mark.xfail
def test_emit_on_with_parent_filters_tables():
    identifier = SimpleNamespace(get_parent_name=lambda: 'a')
    text = 'SELECT * FROM abc a JOIN def d ON a.'
    context = _build_suggest_context('on', text, None, text, identifier)
    assert sorted_dicts(_emit_on(context)) == sorted_dicts([
        {'type': 'column', 'tables': [(None, 'abc', 'a')]},
        # xfail because these currently also are returned
        # {'type': 'table', 'schema': 'a'},
        # {'type': 'view', 'schema': 'a'},
        # {'type': 'function', 'schema': 'a'},
    ])


def test_emit_on_without_parent_uses_fk_join_and_aliases():
    text = 'select a.x, b.y from abc a join bcd b on '
    context = _build_suggest_context('on', text, None, text, empty_identifier())
    assert _emit_on(context) == [
        {'type': 'fk_join', 'tables': [(None, 'abc', 'a'), (None, 'bcd', 'b')]},
        {'type': 'alias', 'aliases': ['a', 'b']},
    ]


def test_emit_on_without_visible_tables_adds_database_and_table():
    text = 'grant select on '
    context = _build_suggest_context('on', text, None, text, empty_identifier())
    assert _emit_on(context) == [
        {'type': 'fk_join', 'tables': []},
        {'type': 'alias', 'aliases': []},
        {'type': 'database'},
        {'type': 'table', 'schema': []},
    ]


def test_emit_database():
    context = _build_suggest_context('database', '', None, '', empty_identifier())
    assert _emit_database(context) == [{'type': 'database'}]


def test_emit_where_token_returns_charset_suggestion_when_available(monkeypatch):
    text = 'select * from tabl where foo = '
    where_token = next(token for token in sqlparse.parse(text)[0].tokens if isinstance(token, sqlparse.sql.Where))
    context = _build_suggest_context(where_token, text, None, text, empty_identifier())
    suggestion = [{'type': 'character_set'}]

    monkeypatch.setattr(completion_engine, '_charset_suggestion', lambda _tokens: suggestion)
    monkeypatch.setattr(
        completion_engine,
        'suggest_based_on_last_token',
        lambda *_args: pytest.fail('suggest_based_on_last_token should not be called'),
    )

    assert _emit_where_token(context) == suggestion


def test_emit_where_token_prepends_enum_value_for_where_fallback(monkeypatch):
    text = 'select * from tabl where foo = '
    where_token = next(token for token in sqlparse.parse(text)[0].tokens if isinstance(token, sqlparse.sql.Where))
    context = _build_suggest_context(where_token, text, None, text, empty_identifier())
    prev_keyword = SimpleNamespace(value='where')
    enum_suggestion = {'type': 'enum_value'}
    fallback = [{'type': 'keyword'}]

    monkeypatch.setattr(completion_engine, '_charset_suggestion', lambda _tokens: None)
    monkeypatch.setattr(completion_engine, 'find_prev_keyword', lambda _text: (prev_keyword, 'select * from tabl where '))
    monkeypatch.setattr(completion_engine, '_enum_value_suggestion', lambda _original, _full: enum_suggestion)
    monkeypatch.setattr(completion_engine, 'suggest_based_on_last_token', lambda *_args: fallback)

    assert _emit_where_token(context) == [enum_suggestion] + fallback


def test_emit_where_token_returns_fallback_for_non_where_keyword(monkeypatch):
    text = 'select * from tabl where foo = '
    where_token = next(token for token in sqlparse.parse(text)[0].tokens if isinstance(token, sqlparse.sql.Where))
    context = _build_suggest_context(where_token, text, None, text, empty_identifier())
    fallback = [{'type': 'keyword'}]

    monkeypatch.setattr(completion_engine, '_charset_suggestion', lambda _tokens: None)
    monkeypatch.setattr(
        completion_engine,
        'find_prev_keyword',
        lambda _text: (SimpleNamespace(value='from'), 'select * from tabl '),
    )
    monkeypatch.setattr(completion_engine, '_enum_value_suggestion', lambda _original, _full: {'type': 'enum_value'})
    monkeypatch.setattr(completion_engine, 'suggest_based_on_last_token', lambda *_args: fallback)

    assert _emit_where_token(context) == fallback


def test_emit_where_token_handles_convert_using_with_trailing_partial_name(monkeypatch):
    text = 'select * from tabl where convert(foo using utf'
    where_token = next(token for token in sqlparse.parse(text)[0].tokens if isinstance(token, sqlparse.sql.Where))
    context = _build_suggest_context(where_token, text, None, text, empty_identifier())

    monkeypatch.setattr(
        completion_engine,
        'suggest_based_on_last_token',
        lambda *_args: pytest.fail('suggest_based_on_last_token should not be called'),
    )

    assert _emit_where_token(context) == [{'type': 'character_set'}]


def test_emit_binary_or_comma_prepends_enum_value_for_where_fallback(monkeypatch):
    text = 'select * from tabl where foo = '
    context = _build_suggest_context('=', text, None, text, empty_identifier())
    prev_keyword = SimpleNamespace(value='where')
    enum_suggestion = {'type': 'enum_value'}
    fallback = [{'type': 'column', 'tables': [(None, 'tabl', None)]}]

    monkeypatch.setattr(completion_engine, 'find_prev_keyword', lambda _text: (prev_keyword, 'select * from tabl where '))
    monkeypatch.setattr(completion_engine, '_enum_value_suggestion', lambda _original, _full: enum_suggestion)
    monkeypatch.setattr(completion_engine, 'suggest_based_on_last_token', lambda *_args: fallback)

    assert _emit_binary_or_comma(context) == [enum_suggestion] + fallback


def test_emit_binary_or_comma_uses_keyword_fallback_for_nonprogressing_rewind(monkeypatch):
    text = 'select * from tabl where foo = '
    context = _build_suggest_context(',', text, None, text, empty_identifier())
    prev_keyword = SimpleNamespace(value='where')
    fallback = [{'type': 'keyword'}]

    monkeypatch.setattr(completion_engine, 'find_prev_keyword', lambda _text: (prev_keyword, text.rstrip()))
    monkeypatch.setattr(completion_engine, '_enum_value_suggestion', lambda _original, _full: None)
    monkeypatch.setattr(
        completion_engine,
        'suggest_based_on_last_token',
        lambda *_args: pytest.fail('suggest_based_on_last_token should not be called'),
    )
    monkeypatch.setattr(completion_engine, '_keyword_suggestions', lambda: fallback)

    assert _emit_binary_or_comma(context) == fallback


def test_emit_binary_or_comma_returns_rewound_fallback_without_where_enum(monkeypatch):
    text = 'select * from tabl and '
    context = _build_suggest_context('and', text, None, text, empty_identifier())
    fallback = [{'type': 'keyword'}]

    monkeypatch.setattr(
        completion_engine,
        'find_prev_keyword',
        lambda _text: (SimpleNamespace(value='from'), 'select * from '),
    )
    monkeypatch.setattr(completion_engine, '_enum_value_suggestion', lambda _original, _full: {'type': 'enum_value'})
    monkeypatch.setattr(completion_engine, 'suggest_based_on_last_token', lambda *_args: fallback)

    assert _emit_binary_or_comma(context) == fallback


@pytest.mark.parametrize(
    ('token', 'expected'),
    [
        (None, False),
        (SimpleNamespace(value='where'), True),
        (SimpleNamespace(value='HAVING'), True),
        (SimpleNamespace(value='from'), False),
        (SimpleNamespace(value=''), False),
    ],
)
def test_is_where_or_having(token, expected):
    assert _is_where_or_having(token) is expected


@pytest.mark.parametrize('exc_type', [TypeError, AttributeError])
def test_suggest_type_returns_keyword_suggestions_when_sqlparse_parse_errors(monkeypatch, exc_type):
    monkeypatch.setattr(completion_engine.sqlparse, 'parse', lambda _text: (_ for _ in ()).throw(exc_type()))

    assert suggest_type('select 1', 'select 1') == [{'type': 'keyword'}]


@pytest.mark.parametrize('exc_type', [TypeError, AttributeError])
def test_suggest_type_returns_keyword_suggestions_when_word_parse_errors(monkeypatch, exc_type):
    parse_inputs: list[str] = []
    original_parse = sqlparse.parse

    def fake_parse(text: str):
        parse_inputs.append(text)
        if len(parse_inputs) == 1:
            return [original_parse('select ')[0]]
        raise exc_type()

    monkeypatch.setattr(completion_engine.sqlparse, 'parse', fake_parse)

    assert suggest_type('select foo', 'select foo') == [{'type': 'keyword'}]
    assert parse_inputs == ['select ', 'foo']


def test_suggest_type_dispatches_backslash_commands_to_suggest_special(monkeypatch):
    parse_inputs: list[str] = []
    special_inputs: list[str] = []
    original_parse = sqlparse.parse

    def fake_parse(text: str):
        parse_inputs.append(text)
        return [original_parse('\\dt ')[0]]

    monkeypatch.setattr(completion_engine.sqlparse, 'parse', fake_parse)
    monkeypatch.setattr(
        completion_engine,
        'suggest_special',
        lambda text: special_inputs.append(text) or [{'type': 'special'}],
    )
    monkeypatch.setattr(
        completion_engine,
        'suggest_based_on_last_token',
        lambda *_args: [{'type': 'keyword'}],
    )

    suggestions = suggest_type('\\dt', '\\dt')

    assert parse_inputs == ['\\dt']
    assert special_inputs == ['\\dt']
    assert suggestions == [{'type': 'special'}]


@pytest.mark.parametrize(
    ('text', 'expected'),
    [
        ('\\', [{'type': 'special'}]),
        ('use ', [{'type': 'database'}]),
        ('connect ', [{'type': 'database'}]),
        ('\\u ', [{'type': 'database'}]),
        ('\\r ', [{'type': 'database'}]),
        ('tableformat ', [{'type': 'table_format'}]),
        ('redirectformat ', [{'type': 'table_format'}]),
        ('\\T ', [{'type': 'table_format'}]),
        ('\\Tr ', [{'type': 'table_format'}]),
        ('\\f ', [{'type': 'favoritequery'}]),
        ('\\fs ', [{'type': 'favoritequery'}]),
        ('\\fd ', [{'type': 'favoritequery'}]),
        ('\\dt ', [{'type': 'table', 'schema': []}, {'type': 'view', 'schema': []}, {'type': 'schema'}]),
        ('\\dt+ ', [{'type': 'table', 'schema': []}, {'type': 'view', 'schema': []}, {'type': 'schema'}]),
        ('\\. ', [{'type': 'file_name'}]),
        ('source ', [{'type': 'file_name'}]),
        ('\\o ', [{'type': 'file_name'}]),
        ('\\once ', [{'type': 'file_name'}]),
        ('tee ', [{'type': 'file_name'}]),
        ('\\e ', [{'type': 'file_name'}]),
        ('\\edit ', [{'type': 'file_name'}]),
        ('\\llm ', [{'type': 'llm'}]),
        ('\\ai ', [{'type': 'llm'}]),
        ('pager ', [{'type': 'keyword'}, {'type': 'special'}]),
    ],
)
def test_suggest_special(text, expected):
    assert suggest_special(text) == expected


@pytest.mark.parametrize(
    ('token', 'text_before_cursor', 'word_before_cursor', 'full_text', 'expected'),
    [
        (None, '', None, '', [{'type': 'keyword'}]),
        ('', '', None, '', [{'type': 'keyword'}, {'type': 'special'}]),
        ('*', 'select *', None, 'select *', [{'type': 'keyword'}]),
        ('as', 'select 1 as ', None, 'select 1 as ', []),
        ('show', 'show ', None, 'show ', [{'type': 'show'}]),
        ('to', 'grant all on db.* to ', None, 'grant all on db.* to ', [{'type': 'user'}]),
        ('to', 'change master to ', None, 'change master to ', [{'type': 'change'}]),
        ('where', 'select * from tabl where ', '9', 'select * from tabl where ', []),
        ('where', 'select * from tabl where "fo', '"fo', 'select * from tabl where "fo', []),
        ('where', "select * from tabl where 'fo", 'fo', "select * from tabl where 'fo", []),
    ],
)
def test_suggest_based_on_last_token(token, text_before_cursor, word_before_cursor, full_text, expected):
    suggestion = suggest_based_on_last_token(
        token,
        text_before_cursor,
        word_before_cursor,
        full_text,
        empty_identifier(),
    )
    assert suggestion == expected


def test_suggest_based_on_last_token_lparen_in_exists_where_suggests_keyword():
    text = 'SELECT * FROM foo WHERE EXISTS ('
    suggestion = suggest_based_on_last_token('(', text, None, text, empty_identifier())
    assert suggestion == [{'type': 'keyword'}]


def test_suggest_based_on_last_token_lparen_in_where_any_suggests_columns_functions():
    text = 'SELECT * FROM tabl WHERE foo = ANY('
    suggestion = suggest_based_on_last_token('(', text, None, text, empty_identifier())
    assert sorted_dicts(suggestion) == sorted_dicts([
        {'type': 'alias', 'aliases': ['tabl']},
        {'type': 'column', 'tables': [(None, 'tabl', None)]},
        {'type': 'function', 'schema': []},
        {'type': 'introducer'},
    ])


def test_suggest_based_on_last_token_lparen_after_join_using_suggests_common_columns():
    text = 'select * from abc inner join def using ('
    suggestion = suggest_based_on_last_token('(', text, None, text, empty_identifier())
    assert suggestion == [{'type': 'column', 'tables': [(None, 'abc', None), (None, 'def', None)], 'drop_unique': True}]


def test_suggest_based_on_last_token_lparen_after_select_subquery_suggests_keyword():
    text = 'SELECT * FROM ('
    suggestion = suggest_based_on_last_token('(', text, None, text, empty_identifier())
    assert suggestion == [{'type': 'keyword'}]


def test_suggest_based_on_last_token_lparen_after_show_suggests_show_items():
    text = 'SHOW ('
    suggestion = suggest_based_on_last_token('(', text, None, text, empty_identifier())
    assert suggestion == [{'type': 'show'}]


def test_suggest_based_on_last_token_lparen_in_function_call_suggests_columns():
    text = 'SELECT MAX('
    full_text = 'SELECT MAX( FROM tbl'
    suggestion = suggest_based_on_last_token('(', text, None, full_text, empty_identifier())
    assert suggestion == [{'type': 'column', 'tables': [(None, 'tbl', None)]}]


@pytest.mark.parametrize(
    ('token', 'text_before_cursor', 'full_text', 'expected'),
    [
        ('call', 'call ', 'call ', [{'type': 'procedure', 'schema': []}]),
        ('set', 'character set', 'character set', [{'type': 'character_set'}]),
        ('distinct', 'select distinct ', 'select distinct ', [{'type': 'column', 'tables': []}]),
        ('database', 'drop database ', 'drop database ', [{'type': 'database'}]),
        ('template', 'create database foo with template ', 'create database foo with template ', [{'type': 'database'}]),
        ('collate', 'collate ', 'collate ', [{'type': 'collation'}]),
        ('table', 'drop table ', 'drop table ', [{'type': 'schema'}, {'type': 'table', 'schema': []}]),
        ('view', 'drop view ', 'drop view ', [{'type': 'schema'}, {'type': 'view', 'schema': []}]),
        ('function', 'drop function ', 'drop function ', [{'type': 'schema'}, {'type': 'function', 'schema': []}]),
    ],
)
def test_suggest_based_on_last_token_direct_keyword_branches(token, text_before_cursor, full_text, expected):
    suggestion = suggest_based_on_last_token(token, text_before_cursor, None, full_text, empty_identifier())
    assert suggestion == expected


def test_suggest_based_on_last_token_relation_keyword_with_schema_parent():
    identifier = SimpleNamespace(get_parent_name=lambda: 'sch')
    text = 'INSERT INTO sch.'
    suggestion = suggest_based_on_last_token('into', text, None, text, identifier)
    assert sorted_dicts(suggestion) == sorted_dicts([
        {'type': 'table', 'schema': 'sch'},
        {'type': 'view', 'schema': 'sch'},
    ])


def test_suggest_based_on_last_token_join_keyword_marks_join_suggestions():
    text = 'SELECT * FROM foo JOIN '
    suggestion = suggest_based_on_last_token(last_non_whitespace_token(text), text, None, text, empty_identifier())
    assert sorted_dicts(suggestion) == sorted_dicts([
        {'type': 'database'},
        {'type': 'table', 'schema': [], 'join': True},
        {'type': 'view', 'schema': []},
    ])


def test_suggest_based_on_last_token_like_in_create_table_suggests_relations():
    text = 'CREATE TABLE new LIKE '
    suggestion = suggest_based_on_last_token('like', text, None, text, empty_identifier())
    assert sorted_dicts(suggestion) == sorted_dicts([
        {'type': 'database'},
        {'type': 'table', 'schema': []},
        {'type': 'view', 'schema': []},
    ])


@pytest.mark.xfail
def test_suggest_based_on_last_token_select_with_parent_identifier_filters_tables():
    identifier = SimpleNamespace(get_parent_name=lambda: 't1')
    text = 'SELECT t1.'
    full_text = 'SELECT t1. FROM tabl1 t1, tabl2 t2'
    suggestion = suggest_based_on_last_token('select', text, None, full_text, identifier)
    assert sorted_dicts(suggestion) == sorted_dicts([
        {'type': 'column', 'tables': [(None, 'tabl1', 't1')]},
        # xfail because these are currently also returned
        # {'type': 'table', 'schema': 't1'},
        # {'type': 'view', 'schema': 't1'},
        # {'type': 'function', 'schema': 't1'},
    ])


def test_suggest_based_on_last_token_select_inside_backticks_adds_keywords():
    text = 'SELECT `a'
    full_text = 'SELECT `a FROM tabl'
    suggestion = suggest_based_on_last_token('select', text, None, full_text, empty_identifier())
    assert sorted_dicts(suggestion) == sorted_dicts([
        {'type': 'column', 'tables': [(None, 'tabl', None)]},
        {'type': 'function', 'schema': []},
        {'type': 'alias', 'aliases': ['tabl']},
        {'type': 'keyword'},
    ])


def test_suggest_based_on_last_token_on_without_parent_suggests_fk_join_and_aliases():
    text = 'select a.x, b.y from abc a join bcd b on '
    suggestion = suggest_based_on_last_token('on', text, None, text, empty_identifier())
    assert suggestion == [
        {'type': 'fk_join', 'tables': [(None, 'abc', 'a'), (None, 'bcd', 'b')]},
        {'type': 'alias', 'aliases': ['a', 'b']},
    ]


def test_suggest_based_on_last_token_on_without_tables_adds_database_and_table():
    text = 'grant select on '
    suggestion = suggest_based_on_last_token('on', text, None, text, empty_identifier())
    assert suggestion == [
        {'type': 'fk_join', 'tables': []},
        {'type': 'alias', 'aliases': []},
        {'type': 'database'},
        {'type': 'table', 'schema': []},
    ]


@pytest.mark.xfail
def test_suggest_based_on_last_token_on_with_parent_identifier_filters_tables():
    identifier = SimpleNamespace(get_parent_name=lambda: 'a')
    text = 'SELECT * FROM abc a JOIN def d ON a.'
    suggestion = suggest_based_on_last_token('on', text, None, text, identifier)
    assert sorted_dicts(suggestion) == sorted_dicts([
        {'type': 'column', 'tables': [(None, 'abc', 'a')]},
        # xfail because these are currently also returned
        # {'type': 'table', 'schema': 'a'},
        # {'type': 'view', 'schema': 'a'},
        # {'type': 'function', 'schema': 'a'},
    ])


def test_suggest_based_on_last_token_binary_operand_in_where_prepends_enum_value():
    text = 'SELECT * FROM tabl WHERE foo = '
    suggestion = suggest_based_on_last_token('=', text, None, text, empty_identifier())
    assert sorted_dicts(suggestion) == sorted_dicts([
        {'type': 'enum_value', 'tables': [(None, 'tabl', None)], 'column': 'foo', 'parent': None},
        {'type': 'alias', 'aliases': ['tabl']},
        {'type': 'column', 'tables': [(None, 'tabl', None)]},
        {'type': 'function', 'schema': []},
        {'type': 'introducer'},
    ])


def test_suggest_based_on_last_token_comma_recurses_to_select_suggestions():
    text = 'SELECT a, '
    full_text = 'SELECT a, FROM tabl'
    suggestion = suggest_based_on_last_token(',', text, None, full_text, empty_identifier())
    assert sorted_dicts(suggestion) == sorted_dicts([
        {'type': 'alias', 'aliases': ['tabl']},
        {'type': 'column', 'tables': [(None, 'tabl', None)]},
        {'type': 'function', 'schema': []},
        {'type': 'introducer'},
    ])


def test_suggest_based_on_last_token_nonprogressing_comma_falls_back_to_keyword():
    text = ','
    suggestion = suggest_based_on_last_token(',', text, None, text, empty_identifier())
    assert suggestion == [{'type': 'keyword'}]


@pytest.mark.parametrize(
    ('identifier', 'schema', 'table', 'alias', 'expected'),
    [
        ('t', None, 'tbl', 't', True),
        ('tbl', None, 'tbl', 't', True),
        ('sch.tbl', 'sch', 'tbl', 't', True),
        ('other', 'sch', 'tbl', 't', False),
        ('sch.other', 'sch', 'tbl', 't', False),
        ('tbl', 'sch', 'other', 't', False),
    ],
)
def test_identifies(identifier, schema, table, alias, expected):
    assert identifies(identifier, schema, table, alias) is expected


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM tabl WHERE foo IN (",
        "SELECT * FROM tabl WHERE foo IN (bar, ",
    ],
)
def test_where_in_suggests_columns(expression):
    suggestions = suggest_type(expression, expression)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["tabl"]},
        {"type": "column", "tables": [(None, "tabl", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_where_equals_any_suggests_columns_or_keywords():
    text = "SELECT * FROM tabl WHERE foo = ANY("
    suggestions = suggest_type(text, text)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["tabl"]},
        {"type": "column", "tables": [(None, "tabl", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_where_convert_using_suggests_character_set():
    text = 'SELECT * FROM tabl WHERE CONVERT(foo USING '
    suggestions = suggest_type(text, text)
    assert suggestions == [{"type": "character_set"}]


def test_where_cast_character_set_suggests_character_set():
    text = 'SELECT * FROM tabl WHERE CAST(foo AS CHAR CHARACTER SET '
    suggestions = suggest_type(text, text)
    assert suggestions == [{"type": "character_set"}]


def test_lparen_suggests_cols():
    suggestion = suggest_type("SELECT MAX( FROM tbl", "SELECT MAX(")
    assert suggestion == [{"type": "column", "tables": [(None, "tbl", None)]}]


def test_operand_inside_function_suggests_cols1():
    suggestion = suggest_type("SELECT MAX(col1 +  FROM tbl", "SELECT MAX(col1 + ")
    assert suggestion == [{"type": "column", "tables": [(None, "tbl", None)]}]


def test_operand_inside_function_suggests_cols2():
    suggestion = suggest_type("SELECT MAX(col1 + col2 +  FROM tbl", "SELECT MAX(col1 + col2 + ")
    assert suggestion == [{"type": "column", "tables": [(None, "tbl", None)]}]


def test_operand_inside_function_suggests_cols3():
    suggestion = suggest_type("SELECT MAX(col1 ||  FROM tbl", "SELECT MAX(col1 || ")
    assert suggestion == [{"type": "column", "tables": [(None, "tbl", None)]}]


def test_operand_inside_function_suggests_cols4():
    suggestion = suggest_type("SELECT MAX(col1 LIKE  FROM tbl", "SELECT MAX(col1 LIKE ")
    assert suggestion == [{"type": "column", "tables": [(None, "tbl", None)]}]


def test_operand_inside_function_suggests_cols5():
    suggestion = suggest_type("SELECT MAX(col1 DIV  FROM tbl", "SELECT MAX(col1 DIV ")
    assert suggestion == [{"type": "column", "tables": [(None, "tbl", None)]}]


@pytest.mark.xfail
def test_arrow_op_inside_function_suggests_nothing():
    suggestion = suggest_type("SELECT MAX(col1->  FROM tbl", "SELECT MAX(col1->")
    assert suggestion == []


def test_select_suggests_cols_and_funcs():
    suggestions = suggest_type("SELECT ", "SELECT ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": []},
        {"type": "column", "tables": []},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM ",
        "INSERT INTO ",
        "COPY ",
        "UPDATE ",
        "DESCRIBE ",
        "DESC ",
        "EXPLAIN ",
    ],
)
def test_expression_suggests_tables_views_and_schemas(expression):
    suggestions = suggest_type(expression, expression)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "table", "schema": []},
        {"type": "view", "schema": []},
        {"type": "database"},
    ])


def test_join_expression_suggests_tables_views_and_schemas():
    expression = "SELECT * FROM foo JOIN "
    suggestions = suggest_type(expression, expression)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "table", "schema": [], "join": True},
        {"type": "view", "schema": []},
        {"type": "database"},
    ])


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM sch.",
        "INSERT INTO sch.",
        "COPY sch.",
        "UPDATE sch.",
        "DESCRIBE sch.",
        "DESC sch.",
        "EXPLAIN sch.",
    ],
)
def test_expression_suggests_qualified_tables_views_and_schemas(expression):
    suggestions = suggest_type(expression, expression)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "table", "schema": "sch"},
        {"type": "view", "schema": "sch"},
    ])


def test_join_expression_suggests_qualified_tables_views_and_schemas():
    expression = "SELECT * FROM foo JOIN sch."
    suggestions = suggest_type(expression, expression)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "table", "schema": "sch", "join": True},
        {"type": "view", "schema": "sch"},
    ])


def test_truncate_suggests_tables_and_schemas():
    suggestions = suggest_type("TRUNCATE ", "TRUNCATE ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "table", "schema": []},
        {"type": "database"},
    ])


def test_truncate_suggests_qualified_tables():
    suggestions = suggest_type("TRUNCATE sch.", "TRUNCATE sch.")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "table", "schema": "sch"},
    ])


def test_distinct_suggests_cols():
    suggestions = suggest_type("SELECT DISTINCT ", "SELECT DISTINCT ")
    assert suggestions == [{"type": "column", "tables": []}]


def test_col_comma_suggests_cols():
    suggestions = suggest_type("SELECT a, b, FROM tbl", "SELECT a, b,")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["tbl"]},
        {"type": "column", "tables": [(None, "tbl", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_table_comma_suggests_tables_and_schemas():
    suggestions = suggest_type("SELECT a, b FROM tbl1, ", "SELECT a, b FROM tbl1, ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "database"},
        {"type": "table", "schema": []},
        {"type": "view", "schema": []},
    ])


def test_into_suggests_tables_and_schemas():
    suggestion = suggest_type("INSERT INTO ", "INSERT INTO ")
    assert sorted_dicts(suggestion) == sorted_dicts([
        {"type": "database"},
        {"type": "table", "schema": []},
        {"type": "view", "schema": []},
    ])


def test_insert_into_lparen_suggests_cols():
    suggestions = suggest_type("INSERT INTO abc (", "INSERT INTO abc (")
    assert suggestions == [{"type": "column", "tables": [(None, "abc", None)]}]


def test_insert_into_lparen_partial_text_suggests_cols():
    suggestions = suggest_type("INSERT INTO abc (i", "INSERT INTO abc (i")
    assert suggestions == [{"type": "column", "tables": [(None, "abc", None)]}]


def test_insert_into_lparen_comma_suggests_cols():
    suggestions = suggest_type("INSERT INTO abc (id,", "INSERT INTO abc (id,")
    assert suggestions == [{"type": "column", "tables": [(None, "abc", None)]}]


def test_partially_typed_col_name_suggests_col_names():
    suggestions = suggest_type("SELECT * FROM tabl WHERE col_n", "SELECT * FROM tabl WHERE col_n")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["tabl"]},
        {"type": "column", "tables": [(None, "tabl", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_dot_suggests_cols_of_a_table_or_schema_qualified_table():
    suggestions = suggest_type("SELECT tabl. FROM tabl", "SELECT tabl.")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "column", "tables": [(None, "tabl", None)]},
        {"type": "table", "schema": "tabl"},
        {"type": "view", "schema": "tabl"},
        {"type": "function", "schema": "tabl"},
    ])


def test_dot_suggests_cols_of_an_alias():
    suggestions = suggest_type("SELECT t1. FROM tabl1 t1, tabl2 t2", "SELECT t1.")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "table", "schema": "t1"},
        {"type": "view", "schema": "t1"},
        {"type": "column", "tables": [(None, "tabl1", "t1")]},
        {"type": "function", "schema": "t1"},
    ])


def test_dot_col_comma_suggests_cols_or_schema_qualified_table():
    suggestions = suggest_type("SELECT t1.a, t2. FROM tabl1 t1, tabl2 t2", "SELECT t1.a, t2.")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "column", "tables": [(None, "tabl2", "t2")]},
        {"type": "table", "schema": "t2"},
        {"type": "view", "schema": "t2"},
        {"type": "function", "schema": "t2"},
    ])


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM (",
        "SELECT * FROM foo WHERE EXISTS (",
        "SELECT * FROM foo WHERE bar AND NOT EXISTS (",
        "SELECT 1 AS",
    ],
)
def test_sub_select_suggests_keyword(expression):
    suggestion = suggest_type(expression, expression)
    assert suggestion == [{"type": "keyword"}]


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM (S",
        "SELECT * FROM foo WHERE EXISTS (S",
        "SELECT * FROM foo WHERE bar AND NOT EXISTS (S",
    ],
)
def test_sub_select_partial_text_suggests_keyword(expression):
    suggestion = suggest_type(expression, expression)
    assert suggestion == [{"type": "keyword"}]


def test_outer_table_reference_in_exists_subquery_suggests_columns():
    q = "SELECT * FROM foo f WHERE EXISTS (SELECT 1 FROM bar WHERE f."
    suggestions = suggest_type(q, q)
    assert suggestions == [
        {"type": "column", "tables": [(None, "foo", "f")]},
        {"type": "table", "schema": "f"},
        {"type": "view", "schema": "f"},
        {"type": "function", "schema": "f"},
    ]


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT * FROM (SELECT * FROM ",
        "SELECT * FROM foo WHERE EXISTS (SELECT * FROM ",
        "SELECT * FROM foo WHERE bar AND NOT EXISTS (SELECT * FROM ",
    ],
)
def test_sub_select_table_name_completion(expression):
    suggestion = suggest_type(expression, expression)
    assert sorted_dicts(suggestion) == sorted_dicts([
        {"type": "database"},
        {"type": "table", "schema": []},
        {"type": "view", "schema": []},
    ])


def test_sub_select_col_name_completion():
    suggestions = suggest_type("SELECT * FROM (SELECT  FROM abc", "SELECT * FROM (SELECT ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["abc"]},
        {"type": "column", "tables": [(None, "abc", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


@pytest.mark.xfail
def test_sub_select_multiple_col_name_completion():
    suggestions = suggest_type("SELECT * FROM (SELECT a, FROM abc", "SELECT * FROM (SELECT a, ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "column", "tables": [(None, "abc", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_sub_select_dot_col_name_completion():
    suggestions = suggest_type("SELECT * FROM (SELECT t. FROM tabl t", "SELECT * FROM (SELECT t.")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "column", "tables": [(None, "tabl", "t")]},
        {"type": "table", "schema": "t"},
        {"type": "view", "schema": "t"},
        {"type": "function", "schema": "t"},
    ])


@pytest.mark.parametrize("join_type", ["", "INNER", "LEFT", "RIGHT OUTER"])
@pytest.mark.parametrize("tbl_alias", ["", "foo"])
def test_join_suggests_tables_and_schemas(tbl_alias, join_type):
    text = f"SELECT * FROM abc {tbl_alias} {join_type} JOIN "
    suggestion = suggest_type(text, text)
    assert sorted_dicts(suggestion) == sorted_dicts([
        {"type": "database"},
        {"type": "table", "schema": [], "join": True},
        {"type": "view", "schema": []},
    ])


@pytest.mark.parametrize(
    "sql",
    [
        "SELECT * FROM abc a JOIN def d ON a.",
        "SELECT * FROM abc a JOIN def d ON a.id = d.id AND a.",
    ],
)
def test_join_alias_dot_suggests_cols1(sql):
    suggestions = suggest_type(sql, sql)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "column", "tables": [(None, "abc", "a")]},
        {"type": "table", "schema": "a"},
        {"type": "view", "schema": "a"},
        {"type": "function", "schema": "a"},
    ])


@pytest.mark.parametrize(
    "sql",
    [
        "SELECT * FROM abc a JOIN def d ON a.id = d.",
        "SELECT * FROM abc a JOIN def d ON a.id = d.id AND a.id2 = d.",
    ],
)
def test_join_alias_dot_suggests_cols2(sql):
    suggestions = suggest_type(sql, sql)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "column", "tables": [(None, "def", "d")]},
        {"type": "table", "schema": "d"},
        {"type": "view", "schema": "d"},
        {"type": "function", "schema": "d"},
    ])


@pytest.mark.parametrize(
    "sql",
    [
        "select a.x, b.y from abc a join bcd b on ",
        "select a.x, b.y from abc a join bcd b on a.id = b.id OR ",
        "select a.x, b.y from abc a join bcd b on a.id = b.id + ",
        "select a.x, b.y from abc a join bcd b on a.id = b.id < ",
    ],
)
def test_on_suggests_aliases(sql):
    suggestions = suggest_type(sql, sql)
    assert suggestions == [
        {"type": "fk_join", "tables": [(None, "abc", "a"), (None, "bcd", "b")]},
        {"type": "alias", "aliases": ["a", "b"]},
    ]


@pytest.mark.parametrize(
    "sql",
    [
        "select abc.x, bcd.y from abc join bcd on ",
        "select abc.x, bcd.y from abc join bcd on abc.id = bcd.id AND ",
    ],
)
def test_on_suggests_tables(sql):
    suggestions = suggest_type(sql, sql)
    assert suggestions == [
        {"type": "fk_join", "tables": [(None, "abc", None), (None, "bcd", None)]},
        {"type": "alias", "aliases": ["abc", "bcd"]},
    ]


@pytest.mark.parametrize(
    "sql",
    [
        "select a.x, b.y from abc a join bcd b on a.id = ",
        "select a.x, b.y from abc a join bcd b on a.id = b.id AND a.id2 = ",
    ],
)
def test_on_suggests_aliases_right_side(sql):
    suggestions = suggest_type(sql, sql)
    assert suggestions == [
        {"type": "fk_join", "tables": [(None, "abc", "a"), (None, "bcd", "b")]},
        {"type": "alias", "aliases": ["a", "b"]},
    ]


@pytest.mark.parametrize(
    "sql",
    [
        "select abc.x, bcd.y from abc join bcd on ",
        "select abc.x, bcd.y from abc join bcd on abc.id = bcd.id and ",
    ],
)
def test_on_suggests_tables_right_side(sql):
    suggestions = suggest_type(sql, sql)
    assert suggestions == [
        {"type": "fk_join", "tables": [(None, "abc", None), (None, "bcd", None)]},
        {"type": "alias", "aliases": ["abc", "bcd"]},
    ]


@pytest.mark.parametrize("col_list", ["", "col1, "])
def test_join_using_suggests_common_columns(col_list):
    text = "select * from abc inner join def using (" + col_list
    assert suggest_type(text, text) == [{"type": "column", "tables": [(None, "abc", None), (None, "def", None)], "drop_unique": True}]


@pytest.mark.parametrize(
    "sql",
    [
        "SELECT * FROM abc a JOIN def d ON a.id = d.id JOIN ghi g ON g.",
        "SELECT * FROM abc a JOIN def d ON a.id = d.id AND a.id2 = d.id2 JOIN ghi g ON d.id = g.id AND g.",
    ],
)
def test_two_join_alias_dot_suggests_cols1(sql):
    suggestions = suggest_type(sql, sql)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "column", "tables": [(None, "ghi", "g")]},
        {"type": "table", "schema": "g"},
        {"type": "view", "schema": "g"},
        {"type": "function", "schema": "g"},
    ])


def test_2_statements_2nd_current():
    suggestions = suggest_type("select * from a; select * from ", "select * from a; select * from ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "table", "schema": []},
        {"type": "view", "schema": []},
        {"type": "database"},
    ])

    suggestions = suggest_type("select * from a; select  from b", "select * from a; select ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["b"]},
        {"type": "column", "tables": [(None, "b", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])

    # Should work even if first statement is invalid
    suggestions = suggest_type("select * from; select * from ", "select * from; select * from ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "table", "schema": []},
        {"type": "view", "schema": []},
        {"type": "database"},
    ])


def test_2_statements_1st_current():
    suggestions = suggest_type("select * from ; select * from b", "select * from ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "database"},
        {"type": "table", "schema": []},
        {"type": "view", "schema": []},
    ])

    suggestions = suggest_type("select  from a; select * from b", "select ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["a"]},
        {"type": "column", "tables": [(None, "a", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_3_statements_2nd_current():
    suggestions = suggest_type("select * from a; select * from ; select * from c", "select * from a; select * from ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "database"},
        {"type": "table", "schema": []},
        {"type": "view", "schema": []},
    ])

    suggestions = suggest_type("select * from a; select  from b; select * from c", "select * from a; select ")
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "alias", "aliases": ["b"]},
        {"type": "column", "tables": [(None, "b", None)]},
        {"type": "function", "schema": []},
        {"type": "introducer"},
    ])


def test_create_db_with_template():
    suggestions = suggest_type("create database foo with template ", "create database foo with template ")

    assert sorted_dicts(suggestions) == sorted_dicts([{"type": "database"}])


@pytest.mark.parametrize("initial_text", ["", "    ", "\t \t"])
def test_specials_included_for_initial_completion(initial_text):
    suggestions = suggest_type(initial_text, initial_text)

    assert sorted_dicts(suggestions) == sorted_dicts([{"type": "keyword"}, {"type": "special"}])


@pytest.mark.parametrize('initial_text', ['REDIRECT'])
def test_specials_included_with_caps(initial_text):
    suggestions = suggest_type(initial_text, initial_text)

    assert sorted_dicts(suggestions) == sorted_dicts([{'type': 'keyword'}, {'type': 'special'}])


def test_specials_not_included_after_initial_token():
    suggestions = suggest_type("create table foo (dt d", "create table foo (dt d")

    assert sorted_dicts(suggestions) == sorted_dicts([{"type": "keyword"}])


def test_drop_schema_qualified_table_suggests_only_tables():
    text = "DROP TABLE schema_name.table_name"
    suggestions = suggest_type(text, text)
    assert suggestions == [{"type": "table", "schema": "schema_name"}]


@pytest.mark.parametrize("text", [",", "  ,", "sel ,"])
def test_handle_pre_completion_comma_gracefully(text):
    suggestions = suggest_type(text, text)

    assert iter(suggestions)


def test_cross_join():
    text = "select * from v1 cross join v2 JOIN v1.id, "
    suggestions = suggest_type(text, text)
    assert sorted_dicts(suggestions) == sorted_dicts([
        {"type": "database"},
        {"type": "table", "schema": [], "join": True},
        {"type": "view", "schema": []},
    ])


@pytest.mark.parametrize(
    "expression",
    [
        "SELECT 1 AS ",
        "SELECT 1 FROM tabl AS ",
    ],
)
def test_after_as(expression):
    suggestions = suggest_type(expression, expression)
    assert set(suggestions) == set()


@pytest.mark.parametrize(
    "expression",
    [
        "\\. ",
        "select 1; \\. ",
        "select 1;\\. ",
        "select 1 ; \\. ",
        "source ",
        "truncate table test; source ",
        "truncate table test ; source ",
        "truncate table test;source ",
    ],
)
def test_source_is_file(expression):
    # "source" has to be registered by hand because that usually happens inside MyCLI in mycli/main.py
    special.register_special_command(
        ...,
        'source',
        '\\. <filename>',
        'Execute commands from file.',
        aliases=[special.SpecialCommandAlias('\\.', case_sensitive=False)],
    )
    suggestions = suggest_type(expression, expression)
    assert suggestions == [{"type": "file_name"}]


@pytest.mark.parametrize(
    "expression",
    [
        "\\f ",
    ],
)
def test_favorite_name_suggestion(expression):
    suggestions = suggest_type(expression, expression)
    assert suggestions == [{"type": "favoritequery"}]


def test_order_by():
    text = "select * from foo order by "
    suggestions = suggest_type(text, text)
    assert suggestions == [{"tables": [(None, "foo", None)], "type": "column"}]


def test_quoted_where():
    text = "'where i=';"
    suggestions = suggest_type(text, text)
    assert suggestions == [{"type": "keyword"}]


def test_find_doubled_backticks_none():
    text = 'select `ab`'
    assert _find_doubled_backticks(text) == []


def test_find_doubled_backticks_some():
    text = 'select `a``b`'
    assert _find_doubled_backticks(text) == [9, 10]


def test_inside_quotes_01():
    text = "select '"
    assert is_inside_quotes(text, len(text)) == 'single'


def test_inside_quotes_02():
    text = "select '\\'"
    assert is_inside_quotes(text, len(text)) == 'single'


def test_inside_quotes_03():
    text = "select '`"
    assert is_inside_quotes(text, len(text)) == 'single'


def test_inside_quotes_04():
    text = 'select "'
    assert is_inside_quotes(text, len(text)) == 'double'


def test_inside_quotes_05():
    text = 'select "\\"\''
    assert is_inside_quotes(text, len(text)) == 'double'


def test_inside_quotes_06():
    text = 'select ""'
    assert is_inside_quotes(text, len(text)) is False


@pytest.mark.parametrize(
    ["text", "position", "expected"],
    [
        ("select `'",      len("select `'"),  'backtick'),
        ("select `' ",     len("select `' "), 'backtick'),
        ("select `'",      -1,  'backtick'),
        ("select `'",      -2,  False),
        ('select `ab` ',   -1,  False),
        ('select `ab` ',   -2,  'backtick'),
        ('select `a``b` ', -1,  False),
        ('select `a``b` ', -2,  'backtick'),
        ('select `a``b` ', -3,  'backtick'),
        ('select `a``b` ', -4,  'backtick'),
        ('select `a``b` ', -5,  'backtick'),
        ('select `a``b` ', -6,  'backtick'),
        ('select `a``b` ', -7,  False),
    ]
)  # fmt: skip
def test_inside_quotes_backtick_01(text, position, expected):
    assert is_inside_quotes(text, position) == expected


def test_inside_quotes_backtick_02():
    """Empty backtick pairs are treated as a doubled (escaped) backtick.
    This is okay because it is invalid SQL, and we don't have to complete on it.
    """
    text = 'select ``'
    assert is_inside_quotes(text, -1) is False


def test_inside_quotes_backtick_03():
    """Empty backtick pairs are treated as a doubled (escaped) backtick.
    This is okay because it is invalid SQL, and we don't have to complete on it.
    """
    text = 'select ``'
    assert is_inside_quotes(text, -2) is False



# type: ignore

"""Steps for behavioral style tests are defined in this module.

Each step is defined by the string decorating it. This string is used
to call the step in "*.feature" file.

"""

from behave import then, when
import wrappers


@when("we refresh completions")
def step_refresh_completions(context):
    """Send refresh command."""
    context.cli.sendline("rehash")


@then('we see text "{text}"')
def step_see_text(context, text):
    """Wait to see given text message."""
    wrappers.expect_exact(context, text, timeout=2)


@then("we see completions refresh started")
def step_see_refresh_started(context):
    """Wait to see refresh output."""
    wrappers.expect_exact(context, "Auto-completion refresh started in the background.", timeout=2)



# type: ignore

from __future__ import annotations

from mycli.packages.special.delimitercommand import DelimiterCommand


def test_delimiter_command_defaults_to_semicolon() -> None:
    command = DelimiterCommand()

    assert command.current == ';'


def test_set_uses_first_argument_token_and_updates_current_delimiter() -> None:
    command = DelimiterCommand()

    result = command.set('$$ select 1 $$')

    assert result[0].status == 'Changed delimiter to $$'
    assert command.current == '$$'


def test_set_rejects_missing_argument() -> None:
    command = DelimiterCommand()

    result = command.set('')

    assert result[0].status == 'Missing required argument, delimiter'
    assert command.current == ';'


def test_set_rejects_delimiter_keyword_case_insensitively() -> None:
    command = DelimiterCommand()

    result = command.set('Delimiter')

    assert result[0].status == 'Invalid delimiter "delimiter"'
    assert command.current == ';'


def test_queries_iter_preserves_statement_text_for_multi_character_delimiter() -> None:
    command = DelimiterCommand()
    command.set('end')

    assert list(command.queries_iter('delete 1end')) == ['delete 1']


def test_queries_iter_with_custom_delimiter_preserves_semicolons_inside_statement() -> None:
    command = DelimiterCommand()
    command.set('$$')

    assert list(command.queries_iter('select 1; select 2$$ select 3$$')) == [
        'select 1; select 2',
        'select 3',
    ]


def test_split_handles_placeholder_collision_in_original_sql() -> None:
    command = DelimiterCommand()
    command.set('$$')

    assert command._split('select \ufffc1; select 2$$ select 3$$') == [
        'select \ufffc1; select 2$$',
        'select 3$$',
    ]


def test_queries_iter_resplits_remaining_input_after_delimiter_change() -> None:
    command = DelimiterCommand()
    queries = command.queries_iter('select 1; delimiter $$ select 2$$ select 3$$')

    assert next(queries) == 'select 1'
    assert next(queries) == 'delimiter $$ select 2$$ select 3$$'

    command.set('$$')

    assert list(queries) == ['select 2', 'select 3']


def test_queries_iter_reappends_old_trailing_delimiter_before_resplitting(monkeypatch) -> None:
    command = DelimiterCommand()
    command._delimiter = ';;'
    split_calls: list[str] = []

    def fake_split(sql: str) -> list[str]:
        split_calls.append(sql)
        if len(split_calls) == 1:
            return ['delimiter $$;;', 'select 2$$']
        return ['ignored', 'select 2']

    monkeypatch.setattr(command, '_split', fake_split)

    queries = command.queries_iter('ignored')

    assert next(queries) == 'delimiter $$'

    command.set('$$')

    assert list(queries) == ['select 2']
    assert split_calls == ['ignored', 'delimiter $$ select 2$$;;']



from types import SimpleNamespace

import click
import pytest

from mycli.packages import interactive_utils


def test_confirm_bool_param_type_converts_bool_and_strings() -> None:
    boolean_type = interactive_utils.ConfirmBoolParamType()

    assert boolean_type.convert(True, None, None) is True
    assert boolean_type.convert(False, None, None) is False
    assert boolean_type.convert('YES', None, None) is True
    assert boolean_type.convert('y', None, None) is True
    assert boolean_type.convert('NO', None, None) is False
    assert boolean_type.convert('n', None, None) is False
    assert repr(boolean_type) == 'BOOL'


def test_confirm_bool_param_type_rejects_invalid_string() -> None:
    boolean_type = interactive_utils.ConfirmBoolParamType()

    with pytest.raises(click.BadParameter, match='maybe is not a valid boolean'):
        boolean_type.convert('maybe', None, None)


def test_confirm_destructive_query_returns_none_when_not_destructive(monkeypatch: pytest.MonkeyPatch) -> None:
    prompt_called = False
    destructive_calls: list[tuple[list[str], str]] = []

    def fake_prompt(*args: object, **kwargs: object) -> bool:
        nonlocal prompt_called
        prompt_called = True
        return True

    def fake_is_destructive(keywords: list[str], query: str) -> bool:
        destructive_calls.append((keywords, query))
        return False

    monkeypatch.setattr(interactive_utils, 'is_destructive', fake_is_destructive)
    monkeypatch.setattr(interactive_utils, 'prompt', fake_prompt)
    monkeypatch.setattr(interactive_utils.sys, 'stdin', SimpleNamespace(isatty=lambda: True))

    keywords = ['drop']
    query = 'select 1;'
    assert interactive_utils.confirm_destructive_query(keywords, query) is None
    assert destructive_calls == [(keywords, query)]
    assert prompt_called is False


def test_confirm_destructive_query_returns_none_without_tty(monkeypatch: pytest.MonkeyPatch) -> None:
    prompt_called = False

    def fake_prompt(*args: object, **kwargs: object) -> bool:
        nonlocal prompt_called
        prompt_called = True
        return True

    monkeypatch.setattr(interactive_utils, 'is_destructive', lambda keywords, query: True)
    monkeypatch.setattr(interactive_utils, 'prompt', fake_prompt)
    monkeypatch.setattr(interactive_utils.sys, 'stdin', SimpleNamespace(isatty=lambda: False))

    keywords = ['drop']
    sql = 'drop database foo;'
    assert interactive_utils.confirm_destructive_query(keywords, sql) is None
    assert prompt_called is False


def test_confirm_destructive_query_prompts_and_returns_user_choice(monkeypatch: pytest.MonkeyPatch) -> None:
    prompt_calls: list[tuple[tuple[object, ...], dict[str, object]]] = []
    destructive_calls: list[tuple[list[str], str]] = []

    def fake_prompt(*args: object, **kwargs: object) -> bool:
        prompt_calls.append((args, dict(kwargs)))
        return True

    def fake_is_destructive(keywords: list[str], query: str) -> bool:
        destructive_calls.append((keywords, query))
        return True

    monkeypatch.setattr(interactive_utils, 'is_destructive', fake_is_destructive)
    monkeypatch.setattr(interactive_utils, 'prompt', fake_prompt)
    monkeypatch.setattr(interactive_utils.sys, 'stdin', SimpleNamespace(isatty=lambda: True))

    keywords = ['drop']
    query = 'drop database foo;'
    result = interactive_utils.confirm_destructive_query(keywords, query)

    assert result is True
    assert destructive_calls == [(keywords, query)]
    assert prompt_calls == [
        (
            ("You're about to run a destructive command.\nDo you want to proceed? (y/n)",),
            {'type': interactive_utils.BOOLEAN_TYPE},
        )
    ]


def test_confirm_destructive_query_returns_false_when_user_rejects(monkeypatch: pytest.MonkeyPatch) -> None:
    prompt_calls: list[tuple[tuple[object, ...], dict[str, object]]] = []
    destructive_calls: list[tuple[list[str], str]] = []

    def fake_prompt(*args: object, **kwargs: object) -> bool:
        prompt_calls.append((args, dict(kwargs)))
        return False

    def fake_is_destructive(keywords: list[str], query: str) -> bool:
        destructive_calls.append((keywords, query))
        return True

    monkeypatch.setattr(interactive_utils, 'is_destructive', fake_is_destructive)
    monkeypatch.setattr(interactive_utils, 'prompt', fake_prompt)
    monkeypatch.setattr(interactive_utils.sys, 'stdin', SimpleNamespace(isatty=lambda: True))

    keywords = ['drop']
    query = 'drop database foo;'
    assert interactive_utils.confirm_destructive_query(keywords, query) is False
    assert destructive_calls == [(keywords, query)]
    assert prompt_calls == [
        (
            ("You're about to run a destructive command.\nDo you want to proceed? (y/n)",),
            {'type': interactive_utils.BOOLEAN_TYPE},
        )
    ]


def test_confirm_returns_false_on_click_abort(monkeypatch: pytest.MonkeyPatch) -> None:
    def fake_confirm(*args: object, **kwargs: object) -> bool:
        raise click.Abort()

    monkeypatch.setattr(click, 'confirm', fake_confirm)

    assert interactive_utils.confirm('continue?') is False


def test_confirm_delegates_to_click_confirm(monkeypatch: pytest.MonkeyPatch) -> None:
    calls: list[tuple[tuple[object, ...], dict[str, object]]] = []

    def fake_confirm(*args: object, **kwargs: object) -> bool:
        calls.append((args, dict(kwargs)))
        return True

    monkeypatch.setattr(click, 'confirm', fake_confirm)

    assert interactive_utils.confirm('continue?', default=True) is True
    assert calls == [(('continue?',), {'default': True})]


def test_prompt_returns_false_on_click_abort(monkeypatch: pytest.MonkeyPatch) -> None:
    def fake_prompt(*args: object, **kwargs: object) -> bool:
        raise click.Abort()

    monkeypatch.setattr(click, 'prompt', fake_prompt)

    assert interactive_utils.prompt('continue?') is False


def test_prompt_delegates_to_click_prompt(monkeypatch: pytest.MonkeyPatch) -> None:
    calls: list[tuple[tuple[object, ...], dict[str, object]]] = []

    def fake_prompt(*args: object, **kwargs: object) -> bool:
        calls.append((args, dict(kwargs)))
        return True

    monkeypatch.setattr(click, 'prompt', fake_prompt)

    assert interactive_utils.prompt('continue?', type=interactive_utils.BOOLEAN_TYPE) is True
    assert calls == [(('continue?',), {'type': interactive_utils.BOOLEAN_TYPE})]



import importlib.util
import os
from pathlib import Path
import platform
import sys
from types import ModuleType
from typing import Any

import pytest

from mycli.packages import filepaths


def load_filepaths_variant(
    monkeypatch: pytest.MonkeyPatch,
    *,
    os_name: str,
    system_name: str,
) -> ModuleType:
    module_path = str(Path(filepaths.__file__).resolve())
    monkeypatch.setattr(os, 'name', os_name, raising=False)
    monkeypatch.setattr(platform, 'system', lambda: system_name)
    module_name = f'filepaths_variant_{os_name}_{system_name}'
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    assert spec is not None
    assert spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


@pytest.mark.skipif(os.name == 'nt', reason='todo: unknown')
def test_default_socket_dirs_import_variants(monkeypatch: pytest.MonkeyPatch) -> None:
    darwin = load_filepaths_variant(monkeypatch, os_name='posix', system_name='Darwin')
    assert darwin.DEFAULT_SOCKET_DIRS == ['/tmp']

    linux = load_filepaths_variant(monkeypatch, os_name='posix', system_name='Linux')
    assert linux.DEFAULT_SOCKET_DIRS == ['/var/run', '/var/lib']

    windows = load_filepaths_variant(monkeypatch, os_name='nt', system_name='Windows')
    assert windows.DEFAULT_SOCKET_DIRS == []


def test_list_path_lists_sql_files_and_directories(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / '.hidden.sql').write_text('select 1\n', encoding='utf-8')
    (tmp_path / 'visible.SQL').write_text('select 1\n', encoding='utf-8')
    (tmp_path / 'notes.txt').write_text('ignored\n', encoding='utf-8')
    (tmp_path / 'folder').mkdir()

    assert filepaths.list_path(str(tmp_path)) == ['visible.SQL', 'folder/']
    assert filepaths.list_path(str(tmp_path / 'missing')) == []


def test_complete_path_and_parse_path() -> None:
    assert filepaths.complete_path('abc', '') == 'abc'
    assert filepaths.complete_path('abcdef', 'abc') == 'abcdef'
    assert filepaths.complete_path('docs', '~') == os.path.join('~', 'docs')
    assert filepaths.complete_path('docs', 'other') == ''

    assert filepaths.parse_path('') == ('', '', 0)
    assert filepaths.parse_path('/tmp/query.sql') == ('/tmp', 'query.sql', -9)
    assert filepaths.parse_path('/tmp/dir/') == ('/tmp/dir', '', 0)


@pytest.mark.skipif(os.name == 'nt', reason='todo: unknown')
def test_suggest_path_branches(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / 'query.sql').write_text('select 1\n', encoding='utf-8')
    (tmp_path / 'subdir').mkdir()

    assert filepaths.suggest_path('') == [
        os.path.abspath(os.sep),
        '~',
        os.curdir,
        os.pardir,
        'query.sql',
        'subdir/',
    ]

    assert filepaths.suggest_path('relative') == ['query.sql', 'subdir/']

    home = tmp_path / 'home'
    home.mkdir()
    (home / 'from_home.sql').write_text('select 1\n', encoding='utf-8')
    monkeypatch.setattr(os.path, 'expanduser', lambda path: str(home))
    assert filepaths.suggest_path('~/f') == ['from_home.sql']

    nested = tmp_path / 'nested'
    nested.mkdir()
    (nested / 'inside.sql').write_text('select 1\n', encoding='utf-8')
    assert filepaths.suggest_path(str(nested / 'missing.sql')) == ['inside.sql']


def test_dir_path_exists(tmp_path: Path) -> None:
    existing = tmp_path / 'logs' / 'mycli.log'
    existing.parent.mkdir()
    assert filepaths.dir_path_exists(str(existing)) is True
    assert filepaths.dir_path_exists(str(tmp_path / 'missing' / 'mycli.log')) is False


@pytest.mark.skipif(os.name == 'nt', reason='todo: unknown')
def test_guess_socket_location_returns_matching_socket(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(filepaths, 'DEFAULT_SOCKET_DIRS', ['/a', '/b'])
    monkeypatch.setattr(filepaths.os.path, 'exists', lambda path: path == '/b')
    monkeypatch.setattr(
        filepaths.os,
        'walk',
        lambda directory, topdown=True: iter([
            ('/b', ['mysql-data', 'other'], ['mysqlx.sock', 'mysql.socket']),
        ]),
    )
    assert filepaths.guess_socket_location() == '/b/mysql.socket'


def test_guess_socket_location_prunes_dirs_and_returns_none(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(filepaths, 'DEFAULT_SOCKET_DIRS', ['/a'])
    monkeypatch.setattr(filepaths.os.path, 'exists', lambda path: True)
    walked_dirs: list[list[str]] = []

    def fake_walk(directory: str, topdown: bool = True) -> Any:
        dirs = ['mysql-data', 'tmp', 'mysqlx', 'other']
        walked_dirs.append(dirs)
        yield (directory, dirs, ['mysqlx.sock', 'readme.txt'])

    monkeypatch.setattr(filepaths.os, 'walk', fake_walk)
    assert filepaths.guess_socket_location() is None
    assert walked_dirs[0] == ['mysql-data', 'mysqlx']



# type: ignore

import os
import pathlib
import tempfile
from unittest.mock import MagicMock

import pymysql
import pytest

import mycli.packages.special.utils
from mycli.packages.special.utils import (
    CACHED_SSL_VERSION,
    format_uptime,
    get_local_timezone,
    get_server_timezone,
    get_ssl_cipher,
    get_ssl_version,
    get_uptime,
    get_warning_count,
    handle_cd_command,
)
from test.utils import TEMPFILE_PREFIX


@pytest.fixture(autouse=True)
def clear_ssl_cache() -> None:
    CACHED_SSL_VERSION.clear()


def test_handle_cd_command_rejects_non_cd_command() -> None:
    handled, message = handle_cd_command(['pwd'])

    assert handled is False
    assert message == 'Not a cd command.'


def test_handle_cd_command_requires_exactly_one_directory() -> None:
    handled, message = handle_cd_command(['cd'])

    assert handled is False
    assert message == 'Exactly one directory name must be provided.'


@pytest.mark.skipif(os.name == 'nt', reason='todo: unknown')
def test_handle_cd_command_changes_directory_and_echoes_cwd(monkeypatch) -> None:
    echoed = []

    monkeypatch.setattr(mycli.packages.special.utils.click, 'echo', lambda message, err=False: echoed.append((message, err)))
    monkeypatch.chdir(os.getcwd())

    # resolve() is needed for mac /private/var arrangement
    with tempfile.TemporaryDirectory(prefix=TEMPFILE_PREFIX) as tempdir:
        tempdir_resolved = str(pathlib.Path(tempdir).resolve())
        handled, message = handle_cd_command(['cd', tempdir_resolved])
        assert str(pathlib.Path(os.getcwd()).resolve()) == tempdir_resolved
        assert handled is True
        assert message is None
        assert echoed == [(tempdir_resolved, True)]


def test_handle_cd_command_returns_oserror_message(monkeypatch) -> None:
    def raise_oserror(directory: str) -> None:
        raise OSError(2, 'No such file or directory')

    monkeypatch.setattr(mycli.packages.special.utils.os, 'chdir', raise_oserror)

    handled, message = handle_cd_command(['cd', '/missing'])

    assert handled is False
    assert message == 'No such file or directory'


def test_format_uptime():
    seconds = 59
    assert '59 sec' == format_uptime(seconds)

    seconds = 120
    assert '2 min 0 sec' == format_uptime(seconds)

    seconds = 54890
    assert '15 hours 14 min 50 sec' == format_uptime(seconds)

    seconds = 598244
    assert '6 days 22 hours 10 min 44 sec' == format_uptime(seconds)

    seconds = 522600
    assert '6 days 1 hour 10 min 0 sec' == format_uptime(seconds)


def test_format_uptime_uses_singular_units() -> None:
    assert format_uptime('90061') == '1 day 1 hour 1 min 1 sec'


def test_get_uptime_returns_value_from_status_row() -> None:
    cur = MagicMock()
    cur.fetchone.return_value = ('Uptime', '15')

    uptime = get_uptime(cur)

    cur.execute.assert_called_once_with('SHOW STATUS LIKE "Uptime"')
    assert uptime == 15


def test_get_uptime_defaults_to_zero_for_missing_value() -> None:
    cur = MagicMock()
    cur.fetchone.return_value = ('Uptime', None)

    assert get_uptime(cur) == 0


def test_get_uptime_ignores_operational_error() -> None:
    cur = MagicMock()
    cur.execute.side_effect = pymysql.err.OperationalError()

    assert get_uptime(cur) == 0


def test_get_warning_count_returns_value_from_count_row() -> None:
    cur = MagicMock()
    cur.fetchone.return_value = ('7',)

    warning_count = get_warning_count(cur)

    cur.execute.assert_called_once_with('SHOW COUNT(*) WARNINGS')
    assert warning_count == 7


def test_get_warning_count_defaults_to_zero_for_missing_value() -> None:
    cur = MagicMock()
    cur.fetchone.return_value = (None,)

    assert get_warning_count(cur) == 0


def test_get_warning_count_ignores_operational_error() -> None:
    cur = MagicMock()
    cur.execute.side_effect = pymysql.err.OperationalError()

    assert get_warning_count(cur) == 0


def test_get_ssl_version_fetches_and_caches_value() -> None:
    cur = MagicMock()
    cur.connection = MagicMock()
    cur.connection.thread_id.return_value = 42
    cur.fetchone.return_value = ('Ssl_version', 'TLSv1.3')

    first = get_ssl_version(cur)
    second = get_ssl_version(cur)

    cur.execute.assert_called_once_with('SHOW STATUS LIKE "Ssl_version"')
    assert first == 'TLSv1.3'
    assert second == 'TLSv1.3'


def test_get_ssl_version_caches_missing_row_as_none() -> None:
    cur = MagicMock()
    cur.connection = MagicMock()
    cur.connection.thread_id.return_value = 42
    cur.fetchone.return_value = None

    first = get_ssl_version(cur)
    second = get_ssl_version(cur)

    cur.execute.assert_called_once_with('SHOW STATUS LIKE "Ssl_version"')
    assert first is None
    assert second is None


def test_get_ssl_version_returns_none_for_empty_value_and_caches_it() -> None:
    cur = MagicMock()
    cur.connection = MagicMock()
    cur.connection.thread_id.return_value = 42
    cur.fetchone.return_value = ('Ssl_version', '')

    first = get_ssl_version(cur)
    second = get_ssl_version(cur)

    cur.execute.assert_called_once_with('SHOW STATUS LIKE "Ssl_version"')
    assert first is None
    assert second is None


def test_get_ssl_version_ignores_operational_error() -> None:
    cur = MagicMock()
    cur.connection = MagicMock()
    cur.connection.thread_id.return_value = 42
    cur.execute.side_effect = pymysql.err.OperationalError()

    assert get_ssl_version(cur) is None


def test_get_ssl_cipher_returns_value() -> None:
    cur = MagicMock()
    cur.fetchone.return_value = ('Ssl_cipher', 'TLS_AES_256_GCM_SHA384')

    ssl_cipher = get_ssl_cipher(cur)

    cur.execute.assert_called_once_with('SHOW STATUS LIKE "Ssl_cipher"')
    assert ssl_cipher == 'TLS_AES_256_GCM_SHA384'


def test_get_ssl_cipher_returns_none_for_missing_row() -> None:
    cur = MagicMock()
    cur.fetchone.return_value = None

    assert get_ssl_cipher(cur) is None


def test_get_ssl_cipher_returns_none_for_empty_value() -> None:
    cur = MagicMock()
    cur.fetchone.return_value = ('Ssl_cipher', '')

    assert get_ssl_cipher(cur) is None


def test_get_ssl_cipher_ignores_operational_error() -> None:
    cur = MagicMock()
    cur.execute.side_effect = pymysql.err.OperationalError()

    assert get_ssl_cipher(cur) is None


def test_get_server_timezone_prefers_system_timezone_when_requested() -> None:
    variables = {
        'time_zone': 'SYSTEM',
        'system_time_zone': 'UTC',
    }

    assert get_server_timezone(variables) == 'UTC'


def test_get_server_timezone_returns_explicit_timezone() -> None:
    variables = {
        'time_zone': '+02:00',
        'system_time_zone': 'UTC',
    }

    assert get_server_timezone(variables) == '+02:00'


def test_get_server_timezone_returns_empty_string_when_keys_are_missing() -> None:
    assert get_server_timezone({}) == ''


def test_get_local_timezone_returns_tzname(monkeypatch) -> None:
    class FakeAwareDatetime:
        def tzname(self) -> str:
            return 'EDT'

    class FakeDatetime:
        @staticmethod
        def now() -> 'FakeDatetime':
            return FakeDatetime()

        def astimezone(self) -> FakeAwareDatetime:
            return FakeAwareDatetime()

    monkeypatch.setattr(mycli.packages.special.utils.datetime, 'datetime', FakeDatetime)

    assert get_local_timezone() == 'EDT'


def test_get_local_timezone_returns_empty_string_when_tzname_is_none(monkeypatch) -> None:
    class FakeAwareDatetime:
        def tzname(self) -> None:
            return None

    class FakeDatetime:
        @staticmethod
        def now() -> 'FakeDatetime':
            return FakeDatetime()

        def astimezone(self) -> FakeAwareDatetime:
            return FakeAwareDatetime()

    monkeypatch.setattr(mycli.packages.special.utils.datetime, 'datetime', FakeDatetime)

    assert get_local_timezone() == ''



# type: ignore

import time
from types import SimpleNamespace
from unittest.mock import Mock, patch

import pytest

import mycli.completion_refresher as completion_refresher


@pytest.fixture
def refresher():
    return completion_refresher.CompletionRefresher()


class FakeThread:
    def __init__(self, target, args, name) -> None:
        self.target = target
        self.args = args
        self.name = name
        self.daemon = False
        self.started = False
        self.alive = False

    def start(self) -> None:
        self.started = True
        self.alive = True

    def run_target(self) -> None:
        try:
            self.target(*self.args)
        finally:
            self.alive = False

    def is_alive(self) -> bool:
        return self.alive


def make_sqlexecute() -> SimpleNamespace:
    return SimpleNamespace(
        dbname='db',
        user='user',
        password='pw',
        host='host',
        port=3306,
        socket='/tmp/mysql.sock',
        character_set='utf8mb4',
        local_infile=False,
        ssl={'ca': 'ca.pem'},
        ssh_user='ssh-user',
        ssh_host='ssh-host',
        ssh_port=22,
        ssh_password='ssh-pw',
        ssh_key_filename='id_rsa',
    )


def test_ctor(refresher) -> None:
    assert len(refresher.refreshers) > 0
    assert list(refresher.refreshers.keys()) == [
        "databases",
        "schemata",
        "tables",
        "foreign_keys",
        "enum_values",
        "users",
        "functions",
        "procedures",
        'character_sets',
        'collations',
        "special_commands",
        "show_commands",
        "keywords",
    ]


def test_refresh_called_once(refresher):
    """

    :param refresher:
    :return:
    """
    callbacks = Mock()
    sqlexecute = Mock()

    with patch.object(refresher, "_bg_refresh") as bg_refresh:
        actual = refresher.refresh(sqlexecute, callbacks)
        time.sleep(1)  # Wait for the thread to work.
        assert actual[0].preamble is None
        assert actual[0].header is None
        assert actual[0].rows is None
        assert actual[0].status == "Auto-completion refresh started in the background."
        bg_refresh.assert_called_with(sqlexecute, callbacks, {})


def test_refresh_called_twice(refresher):
    """If refresh is called a second time, it should be restarted.

    :param refresher:
    :return:

    """
    callbacks = Mock()

    sqlexecute = Mock()

    def dummy_bg_refresh(*args):
        time.sleep(3)  # seconds

    refresher._bg_refresh = dummy_bg_refresh

    actual1 = refresher.refresh(sqlexecute, callbacks)
    time.sleep(1)  # Wait for the thread to work.
    assert actual1[0].preamble is None
    assert actual1[0].header is None
    assert actual1[0].rows is None
    assert actual1[0].status == "Auto-completion refresh started in the background."

    actual2 = refresher.refresh(sqlexecute, callbacks)
    time.sleep(1)  # Wait for the thread to work.
    assert actual2[0].preamble is None
    assert actual2[0].header is None
    assert actual2[0].rows is None
    assert actual2[0].status == "Auto-completion refresh restarted."
    assert refresher._completer_thread is not None
    refresher._completer_thread.join()


def test_refresh_with_callbacks(refresher):
    """Callbacks must be called.

    :param refresher:

    """
    callbacks = [Mock()]
    sqlexecute_class = Mock()
    sqlexecute = Mock()

    with patch("mycli.completion_refresher.SQLExecute", sqlexecute_class):
        # Set refreshers to 0: we're not testing refresh logic here
        refresher.refreshers = {}
        refresher.refresh(sqlexecute, callbacks)
        time.sleep(1)  # Wait for the thread to work.
        assert callbacks[0].call_count == 1


def test_refresh_starts_background_thread(monkeypatch, refresher) -> None:
    calls: list[tuple[object, object, dict]] = []

    def fake_bg_refresh(executor, callbacks, options) -> None:
        calls.append((executor, callbacks, options))

    monkeypatch.setattr(completion_refresher.threading, 'Thread', FakeThread)
    monkeypatch.setattr(refresher, '_bg_refresh', fake_bg_refresh)

    sqlexecute = Mock()
    callbacks = Mock()

    actual = refresher.refresh(sqlexecute, callbacks)

    assert actual[0].status == "Auto-completion refresh started in the background."
    assert refresher._completer_thread is not None
    assert refresher._completer_thread.name == "completion_refresh"
    assert refresher._completer_thread.daemon is True
    assert refresher._completer_thread.started is True
    assert refresher.is_refreshing() is True
    assert calls == []

    refresher._completer_thread.run_target()
    assert calls == [(sqlexecute, callbacks, {})]
    assert refresher.is_refreshing() is False


def test_refresh_passes_explicit_completer_options(monkeypatch, refresher) -> None:
    calls: list[tuple[object, object, dict]] = []

    def fake_bg_refresh(executor, callbacks, options) -> None:
        calls.append((executor, callbacks, options))

    monkeypatch.setattr(completion_refresher.threading, 'Thread', FakeThread)
    monkeypatch.setattr(refresher, '_bg_refresh', fake_bg_refresh)

    sqlexecute = Mock()
    callbacks = Mock()
    options = {'smart_completion': True}

    refresher.refresh(sqlexecute, callbacks, options)
    refresher._completer_thread.run_target()

    assert calls == [(sqlexecute, callbacks, options)]


def test_refresh_while_refreshing_restarts(monkeypatch, refresher) -> None:
    thread_calls: list[tuple[object, object, str]] = []

    def fail_thread(*, target, args, name):
        thread_calls.append((target, args, name))
        return FakeThread(target, args, name)

    monkeypatch.setattr(completion_refresher.threading, 'Thread', fail_thread)
    existing_thread = SimpleNamespace(is_alive=lambda: True)
    refresher._completer_thread = existing_thread

    actual = refresher.refresh(Mock(), Mock())

    assert actual[0].status == "Auto-completion refresh restarted."
    assert refresher._restart_refresh.is_set() is True
    assert refresher._completer_thread is existing_thread
    assert thread_calls == []


def test_bg_refresh_restarts_wraps_callbacks_and_closes(monkeypatch, refresher) -> None:
    completers: list[SimpleNamespace] = []
    executor_inits: list[tuple[object, ...]] = []
    executors: list[object] = []
    refresher_calls: list[str] = []
    callback_calls: list[tuple[str, SimpleNamespace]] = []
    event_order: list[str] = []

    class FakeCompleter:
        tidb_functions = ['tidb-func']
        tidb_keywords = ['tidb-keyword']

        def __init__(self, **options) -> None:
            self.options = options
            completers.append(self)

    class FakeExecutor:
        def __init__(self, *args) -> None:
            executor_inits.append(args)
            self.closed = False
            executors.append(self)

        def close(self) -> None:
            self.closed = True
            event_order.append('close')

    def first_refresher(completer, executor) -> None:
        refresher_calls.append('first')
        event_order.append('refresher:first')
        if refresher_calls == ['first']:
            refresher._restart_refresh.set()

    def second_refresher(completer, executor) -> None:
        refresher_calls.append('second')
        event_order.append('refresher:second')

    def first_callback(completer) -> None:
        callback_calls.append(('first', completer))
        event_order.append('callback:first')

    def second_callback(completer) -> None:
        callback_calls.append(('second', completer))
        event_order.append('callback:second')

    monkeypatch.setattr(completion_refresher, 'SQLCompleter', FakeCompleter)
    monkeypatch.setattr(completion_refresher, 'SQLExecute', FakeExecutor)
    refresher.refreshers = {
        'first': first_refresher,
        'second': second_refresher,
    }

    sqlexecute = make_sqlexecute()
    refresher._bg_refresh(sqlexecute, [first_callback, second_callback], {'smart_completion': True})

    assert len(completers) == 1
    assert completers[0].options == {'smart_completion': True}
    assert executor_inits == [
        (
            'db',
            'user',
            'pw',
            'host',
            3306,
            '/tmp/mysql.sock',
            'utf8mb4',
            False,
            {'ca': 'ca.pem'},
            'ssh-user',
            'ssh-host',
            22,
            'ssh-pw',
            'id_rsa',
        )
    ]
    assert len(executors) == 1
    assert executors[0].closed is True
    assert refresher_calls == ['first', 'first', 'second']
    assert refresher._restart_refresh.is_set() is False
    assert callback_calls == [('first', completers[0]), ('second', completers[0])]
    assert event_order == [
        'refresher:first',
        'refresher:first',
        'refresher:second',
        'callback:first',
        'callback:second',
        'close',
    ]


def test_bg_refresh_wraps_single_callback_callable(monkeypatch, refresher) -> None:
    completers: list[SimpleNamespace] = []

    class FakeCompleter:
        tidb_functions = []
        tidb_keywords = []

        def __init__(self, **options) -> None:
            completers.append(self)

    class FakeExecutor:
        def __init__(self, *args) -> None:
            self.closed = False

        def close(self) -> None:
            self.closed = True

    callback = Mock()

    monkeypatch.setattr(completion_refresher, 'SQLCompleter', FakeCompleter)
    monkeypatch.setattr(completion_refresher, 'SQLExecute', FakeExecutor)
    refresher.refreshers = {}

    refresher._bg_refresh(make_sqlexecute(), callback, {})

    callback.assert_called_once_with(completers[0])


def test_refresher_decorator_registers_function() -> None:
    refreshers: dict[str, object] = {}

    @completion_refresher.refresher('demo', refreshers=refreshers)
    def demo(completer, executor) -> None:
        return None

    assert refreshers == {'demo': demo}


def test_refresh_helpers_delegate_to_completer_and_executor(monkeypatch) -> None:
    completer = Mock()
    executor = Mock()
    executor.dbname = 'current_db'
    executor.databases.return_value = ['db1', 'db2']
    executor.table_columns.return_value = iter([('tbl', 'col')])
    executor.foreign_keys.return_value = iter([('tbl', 'col', 'other', 'id')])
    executor.enum_values.return_value = iter([('tbl', 'status', ['open'])])
    executor.users.return_value = iter([('app',)])
    executor.procedures.return_value = iter([('proc',)])
    executor.character_sets.return_value = iter([('utf8mb4',)])
    executor.collations.return_value = iter([('utf8mb4_unicode_ci',)])
    executor.show_candidates.return_value = iter([('FULL TABLES',)])

    monkeypatch.setattr(completion_refresher, 'COMMANDS', {'\\x': object(), 'help': object()})

    completion_refresher.refresh_databases(completer, executor)
    completion_refresher.refresh_schemata(completer, executor)
    completion_refresher.refresh_tables(completer, executor)
    completion_refresher.refresh_foreign_keys(completer, executor)
    completion_refresher.refresh_enum_values(completer, executor)
    completion_refresher.refresh_users(completer, executor)
    completion_refresher.refresh_procedures(completer, executor)
    completion_refresher.refresh_character_sets(completer, executor)
    completion_refresher.refresh_collations(completer, executor)
    completion_refresher.refresh_special(completer, executor)
    completion_refresher.refresh_show_commands(completer, executor)

    completer.extend_database_names.assert_called_once_with(['db1', 'db2'])
    completer.extend_schemata.assert_called_once_with('current_db')
    completer.set_dbname.assert_called_once_with('current_db')
    completer.extend_relations.assert_called_once_with([('tbl', 'col')], kind='tables')
    completer.extend_columns.assert_called_once_with([('tbl', 'col')], kind='tables')
    completer.extend_foreign_keys.assert_called_once_with(executor.foreign_keys.return_value)
    completer.extend_enum_values.assert_called_once_with(executor.enum_values.return_value)
    completer.extend_users.assert_called_once_with(executor.users.return_value)
    completer.extend_procedures.assert_called_once_with(executor.procedures.return_value)
    completer.extend_character_sets.assert_called_once_with(executor.character_sets.return_value)
    completer.extend_collations.assert_called_once_with(executor.collations.return_value)
    completer.extend_special_commands.assert_called_once_with(['\\x', 'help'])
    completer.extend_show_items.assert_called_once_with(executor.show_candidates.return_value)


def test_refresh_functions_extends_tidb_builtins_only_for_tidb() -> None:
    completer = Mock()
    completer.tidb_functions = ['tidb_func']

    executor = Mock()
    executor.functions.return_value = iter([('func',)])
    executor.server_info = SimpleNamespace(species=completion_refresher.ServerSpecies.TiDB)

    completion_refresher.refresh_functions(completer, executor)

    assert completer.extend_functions.call_args_list == [
        ((executor.functions.return_value,), {}),
        ((['tidb_func'],), {'builtin': True}),
    ]

    completer.reset_mock()
    executor.server_info = SimpleNamespace(species=completion_refresher.ServerSpecies.MySQL)

    completion_refresher.refresh_functions(completer, executor)

    assert completer.extend_functions.call_args_list == [
        ((executor.functions.return_value,), {}),
    ]

    completer.reset_mock()
    executor.server_info = None

    completion_refresher.refresh_functions(completer, executor)

    assert completer.extend_functions.call_args_list == [
        ((executor.functions.return_value,), {}),
    ]


def test_refresh_keywords_extends_tidb_keywords_only_for_tidb() -> None:
    completer = Mock()
    completer.tidb_keywords = ['FLASHBACK']

    executor = Mock()
    executor.server_info = SimpleNamespace(species=completion_refresher.ServerSpecies.TiDB)

    completion_refresher.refresh_keywords(completer, executor)

    completer.extend_keywords.assert_called_once_with(['FLASHBACK'], replace=True)

    completer.reset_mock()
    executor.server_info = SimpleNamespace(species=completion_refresher.ServerSpecies.MySQL)

    completion_refresher.refresh_keywords(completer, executor)

    completer.extend_keywords.assert_not_called()

    completer.reset_mock()
    executor.server_info = None

    completion_refresher.refresh_keywords(completer, executor)

    completer.extend_keywords.assert_not_called()



from __future__ import annotations

import builtins
import importlib
from pathlib import Path
import sys
from typing import TextIO

import pytest

from mycli.packages import paramiko_stub, ssh_utils


class FakeSSHConfig:
    def __init__(self, parse_error: Exception | None = None) -> None:
        self.parse_error = parse_error
        self.parsed_text: str | None = None

    def parse(self, handle: TextIO) -> None:
        if self.parse_error is not None:
            raise self.parse_error
        self.parsed_text = handle.read()


def test_read_ssh_config_parses_and_returns_config(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    config_path = tmp_path / 'ssh_config'
    config_path.write_text('Host demo\n  HostName example.com\n', encoding='utf-8')
    fake_ssh_config = FakeSSHConfig()

    monkeypatch.setattr(ssh_utils.paramiko.config, 'SSHConfig', lambda: fake_ssh_config)

    result = ssh_utils.read_ssh_config(str(config_path))

    assert result is fake_ssh_config
    assert fake_ssh_config.parsed_text == 'Host demo\n  HostName example.com\n'


def test_read_ssh_config_reports_missing_file_and_exits(monkeypatch: pytest.MonkeyPatch) -> None:
    secho_calls: list[tuple[str, bool, str]] = []

    monkeypatch.setattr(
        ssh_utils.click,
        'secho',
        lambda message, err, fg: secho_calls.append((message, err, fg)),
    )

    with pytest.raises(SystemExit) as excinfo:
        ssh_utils.read_ssh_config('/definitely/missing/ssh_config')

    assert excinfo.value.code == 1
    assert secho_calls == [("[Errno 2] No such file or directory: '/definitely/missing/ssh_config'", True, 'red')]


def test_read_ssh_config_reports_parse_errors_and_exits(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    config_path = tmp_path / 'ssh_config'
    config_path.write_text('Host broken\n', encoding='utf-8')
    fake_ssh_config = FakeSSHConfig(parse_error=RuntimeError('bad config'))
    secho_calls: list[tuple[str, bool, str]] = []

    monkeypatch.setattr(ssh_utils.paramiko.config, 'SSHConfig', lambda: fake_ssh_config)
    monkeypatch.setattr(
        ssh_utils.click,
        'secho',
        lambda message, err, fg: secho_calls.append((message, err, fg)),
    )

    with pytest.raises(SystemExit) as excinfo:
        ssh_utils.read_ssh_config(str(config_path))

    assert excinfo.value.code == 1
    assert secho_calls == [(f'Could not parse SSH configuration file {config_path}:\nbad config ', True, 'red')]


def test_ssh_utils_falls_back_to_paramiko_stub_when_paramiko_is_unavailable(monkeypatch: pytest.MonkeyPatch) -> None:
    original_import = builtins.__import__

    def fake_import(name: str, globals_=None, locals_=None, fromlist=(), level: int = 0):
        if name == 'paramiko':
            raise ImportError('paramiko not installed')
        return original_import(name, globals_, locals_, fromlist, level)

    monkeypatch.delitem(sys.modules, 'paramiko', raising=False)
    monkeypatch.setattr(builtins, '__import__', fake_import)

    reloaded = importlib.reload(ssh_utils)

    assert reloaded.paramiko is paramiko_stub.paramiko

    monkeypatch.undo()
    importlib.reload(ssh_utils)



import importlib.metadata
import json
from types import SimpleNamespace
import urllib.error

from mycli.main_modes import checkup


class FakeUrlResponse:
    def __init__(self, payload: dict) -> None:
        self.payload = payload

    def __enter__(self) -> 'FakeUrlResponse':
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        return None

    def read(self) -> bytes:
        return json.dumps(self.payload).encode('utf8')


def test_pypi_api_fetch_success(monkeypatch) -> None:
    def fake_urlopen(url: str, timeout: int) -> FakeUrlResponse:
        assert url == 'https://pypi.org/pypi/mycli/json'
        assert timeout == 5
        return FakeUrlResponse({'info': {'version': '1.2.3'}})

    monkeypatch.setattr(checkup.urllib.request, 'urlopen', fake_urlopen)

    assert checkup.pypi_api_fetch('/mycli/json') == {'info': {'version': '1.2.3'}}


def test_pypi_api_fetch_url_error(monkeypatch, capsys) -> None:
    def fake_urlopen(url: str, timeout: int) -> FakeUrlResponse:
        raise urllib.error.URLError('offline')

    monkeypatch.setattr(checkup.urllib.request, 'urlopen', fake_urlopen)

    assert checkup.pypi_api_fetch('mycli/json') == {}
    assert 'Failed to connect to PyPi on https://pypi.org/pypi/mycli/json' in capsys.readouterr().err


def test_dependencies_checkup(monkeypatch, capsys) -> None:
    versions = {
        'cli_helpers': '1.0.0',
        'click': '2.0.0',
        'prompt_toolkit': '3.0.0',
        'pymysql': '4.0.0',
        'pygments': '2.19.2',
        'sqlglot': '30.7.0',
        'sqlglotc': '30.7.0',
    }

    def fake_version(name: str) -> str:
        if name == 'tabulate':
            raise importlib.metadata.PackageNotFoundError
        return versions[name]

    def fake_pypi_api_fetch(fragment: str) -> dict:
        dependency = fragment.strip('/').removesuffix('/json')
        return {'info': {'version': f'latest-{dependency}'}}

    monkeypatch.setattr(checkup.importlib.metadata, 'version', fake_version)
    monkeypatch.setattr(checkup, 'pypi_api_fetch', fake_pypi_api_fetch)

    checkup._dependencies_checkup()
    output = capsys.readouterr().out

    assert '### Key Python dependencies:' in output
    assert 'cli_helpers version 1.0.0 (latest latest-cli_helpers)' in output
    assert 'click version 2.0.0 (latest latest-click)' in output
    assert 'prompt_toolkit version 3.0.0 (latest latest-prompt_toolkit)' in output
    assert 'pymysql version 4.0.0 (latest latest-pymysql)' in output
    assert 'tabulate version None (latest latest-tabulate)' in output


def test_executables_checkup(monkeypatch, capsys) -> None:
    monkeypatch.setattr(
        checkup.shutil,
        'which',
        lambda executable: f'/usr/bin/{executable}' if executable != 'fzf' else None,
    )

    checkup._executables_checkup()
    output = capsys.readouterr().out

    assert '### External executables:' in output
    assert 'The "less" executable was found' in output
    assert 'The recommended "fzf" executable was not found' in output
    assert 'The "pygmentize" executable was found' in output


def test_environment_checkup(monkeypatch, capsys) -> None:
    monkeypatch.setenv('EDITOR', 'vim')
    monkeypatch.delenv('VISUAL', raising=False)

    checkup._environment_checkup()
    output = capsys.readouterr().out

    assert '### Environment variables:' in output
    assert 'The $EDITOR environment variable was set to "vim" ' in output
    assert 'The $VISUAL environment variable was not set' in output


def test_configuration_checkup_missing_file(capsys) -> None:
    mycli = SimpleNamespace(
        config={},
        config_without_package_defaults={},
        config_without_user_options={},
    )

    checkup._configuration_checkup(mycli)
    output = capsys.readouterr().out

    assert '### Missing file:' in output
    assert 'The local ~/,myclirc is missing or empty.' in output
    assert f'{checkup.REPO_URL}/blob/main/mycli/myclirc' in output


def test_configuration_checkup_reports_missing_unsupported_and_deprecated(capsys) -> None:
    mycli = SimpleNamespace(
        config={
            'main': {
                'present': '',
                'missing_item': '',
            },
            'extra_section': {
                'extra_item': '',
            },
        },
        config_without_package_defaults={
            'main': {
                'present': '',
                'unsupported_item': '',
                'default_character_set': '',
            },
            'unsupported_section': {
                'anything': '',
            },
            'colors': {
                'sql.keyword': '',
            },
            'favorite_queries': {
                'demo': 'select 1',
            },
        },
        config_without_user_options={
            'main': {
                'present': '',
            },
            'colors': {},
        },
    )

    checkup._configuration_checkup(mycli)
    output = capsys.readouterr().out

    assert '### Missing in user ~/.myclirc:' in output
    assert 'The entire section:\n\n    [extra_section]\n' in output
    assert 'The item:\n\n    [main]\n    missing_item =' in output
    assert '### Unsupported in user ~/.myclirc:' in output
    assert 'The entire section:\n\n    [unsupported_section]\n' in output
    assert 'The item:\n\n    [main]\n    unsupported_item =' in output
    assert '### Deprecated in user ~/.myclirc:' in output
    assert '    [main]\n    default_character_set' in output
    assert '    [connection]\n    default_character_set' in output
    assert f'{checkup.REPO_URL}/blob/main/mycli/myclirc' in output


def test_configuration_checkup_skips_transitioned_and_free_entry_items(capsys) -> None:
    mycli = SimpleNamespace(
        config={
            'extra_section': {
                'extra_item': '',
            },
            'connection': {
                'default_character_set': '',
            },
        },
        config_without_package_defaults={
            'connection': {},
            'unsupported_section': {
                'anything': '',
            },
            'favorite_queries': {
                'demo': 'select 1',
            },
        },
        config_without_user_options={
            'connection': {},
            'favorite_queries': {},
        },
    )

    checkup._configuration_checkup(mycli)
    output = capsys.readouterr().out

    assert 'Missing in user ~/.myclirc:' in output
    assert 'The entire section:\n\n    [extra_section]\n' in output
    assert 'Unsupported in user ~/.myclirc:' in output
    assert 'The entire section:\n\n    [unsupported_section]\n' in output
    assert '[connection]\n    default_character_set =' not in output
    assert '[favorite_queries]' not in output


def test_configuration_checkup_up_to_date(capsys) -> None:
    mycli = SimpleNamespace(
        config={
            'main': {
                'prompt': '',
            },
        },
        config_without_package_defaults={
            'main': {
                'prompt': '',
            },
        },
        config_without_user_options={
            'main': {
                'prompt': '',
            },
        },
    )

    checkup._configuration_checkup(mycli)
    output = capsys.readouterr().out

    assert '### Configuration:' in output
    assert 'User configuration all up to date!' in output


def test_main_checkup_calls_all_sections(monkeypatch) -> None:
    calls: list[tuple[str, object]] = []
    mycli = SimpleNamespace(name='mycli')

    monkeypatch.setattr(checkup, '_dependencies_checkup', lambda: calls.append(('dependencies', None)))
    monkeypatch.setattr(checkup, '_executables_checkup', lambda: calls.append(('executables', None)))
    monkeypatch.setattr(checkup, '_environment_checkup', lambda: calls.append(('environment', None)))
    monkeypatch.setattr(checkup, '_configuration_checkup', lambda arg: calls.append(('configuration', arg)))

    checkup.main_checkup(mycli)

    assert calls == [
        ('dependencies', None),
        ('executables', None),
        ('environment', None),
        ('configuration', mycli),
    ]
